# -*- coding: utf-8 -*-

# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html
from stock10jqka.items import Stock10JqkaItem, NewsItem, StockInfoItem
from stock10jqka.settings import COLUMNS_STOCK, COLUMNS_NEW, COLUMNS_STOCK_ITEM
from yangke.common.mysql import insert_item_to_mysql, has_table, create_table, update_update_time_of_table


class Stock10JqkaPipeline(object):
    """
    爬虫中yield的item会送到该类中执行，调用该类中的process_item()方法
    """

    def __init__(self):
        pass

    def open_spider(self, spider):
        # 确保mysql服务开启
        print("开始爬取数据...")

    def process_item(self, item, spider):
        if isinstance(item, Stock10JqkaItem):
            # 保存数据
            print("插入股票数据，股票代码：{}".format(item['code']))
            tableName = "ths{}".format(item['code'])
            if not has_table(tableName):  # 这句实际不需要判断，因为创建表时，只有表不存在时才会创建
                create_table(tableName, COLUMNS_STOCK)  # 如果表存在时调用该函数，会有warning提示重复

            item_dict = item.deepcopy()
            item_dict.pop('code')
            ITEM_TO_COLUMNS_STOCK = dict(zip(COLUMNS_STOCK_ITEM.values(), COLUMNS_STOCK_ITEM.keys()))  # 反转字典
            COLUMNS_LIST = [ITEM_TO_COLUMNS_STOCK.get(key) for key in item_dict.keys()]  # 逐个获得item对应的mysql列名
            insert_item_to_mysql(table_name=tableName, values=list(item_dict.values()),
                                 col_names=COLUMNS_LIST, ignore=True)
            update_update_time_of_table(tableName)
        elif isinstance(item, NewsItem):
            # 保存新闻
            if item['date'] is None and item['标题'] is None:  # 有些重定向的页面链接到了同花顺以外的域名，且获取不到数据
                return
            print("插入股票新闻，股票代码：{}".format(item['code']))
            tableName = "news{}".format(item['code'])
            item_dict = item.deepcopy()
            item_dict.pop('code')
            if not has_table(table_name=tableName):
                create_table(tableName, COLUMNS_NEW, primary=[0, 1])
            insert_item_to_mysql(table_name=tableName, values=list(item_dict.values()),
                                 col_names=(list(item_dict.keys())), replace=True)
            update_update_time_of_table(tableName)
        # 将item保存到mysql
        elif isinstance(item, StockInfoItem):  # 记录各个股票的基本信息
            if not has_table(table_name="stocksInfo"):
                create_table("stocksInfo", columns={"symbol": "varchar(6)", "business": "varchar(1000)",
                                                    "concept": "varchar(100)", "date_listing": "date",
                                                    "location": "varchar(100)", "name": "varchar(20)"}, primary=[0])
            insert_item_to_mysql(table_name="stocksInfo", col_names=["symbol", "business", "concept", "date_listing",
                                                                     "location", "name"],
                                 values=[item['code'], item['business'], item['concept'], item['date_listing'],
                                         item['location'], item['name']], replace=True)
            update_update_time_of_table("stocksInfo")
            # 构建neo4j数据库

        return item

    def close_spider(self, spider):
        print("爬取结束！")


def get_cursor_news():
    global cursor_news
    if cursor_news is None:
        return Stock10JqkaPipeline().cursor_news
    else:
        return cursor_news


def get_cursor_stocks():
    global cursor_stocks
    if cursor_stocks is None:
        cursor_stocks = Stock10JqkaPipeline().cursor_stocks
    return cursor_stocks



# -*- coding: utf-8 -*-

# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

import scrapy


class Stock10JqkaItem(scrapy.Item):
    """
    该数据item存储到Mysql数据库，这里item的字段需要与settings中COLUMN_STOCK的字段相同
    """
    # define the fields for your item here like:
    # name = scrapy.Field()
    code = scrapy.Field()  # 代码，唯一标识符，表示数据属于哪个股票
    date = scrapy.Field()
    诊股_评分 = scrapy.Field()
    诊股_击败 = scrapy.Field()
    诊股_短期趋势 = scrapy.Field()
    诊股_中期趋势 = scrapy.Field()
    诊股_长期趋势 = scrapy.Field()
    诊股_技术面 = scrapy.Field()
    诊股_资金面 = scrapy.Field()
    诊股_消息面 = scrapy.Field()
    诊股_行业面 = scrapy.Field()
    诊股_基本面 = scrapy.Field()

    评级 = scrapy.Field()

    # 历史数据
    # 因为所属地域，主营业务等可能会有改变，这里把历史数据也存储下来
    地域 = scrapy.Field()  # 所属地域，当前的
    设计概念 = scrapy.Field()  # 设计概念
    主营业务 = scrapy.Field()  # 主营业务
    每股净资产 = scrapy.Field()
    每股收益 = scrapy.Field()
    净利润 = scrapy.Field()
    净利润增长率 = scrapy.Field()
    营业收入 = scrapy.Field()
    每股现金流 = scrapy.Field()
    每股公积金 = scrapy.Field()
    每股未分配利润 = scrapy.Field()
    总股本 = scrapy.Field()
    流通股 = scrapy.Field()


class StockInfoItem(scrapy.Item):
    """
    该数据item用于构建neo4j图数据

    字段：code, name, location, concept, business, date_listing
    """

    code = scrapy.Field()  # 代码
    name = scrapy.Field()  # 名称
    location = scrapy.Field()  # 所属地域，当前的
    concept = scrapy.Field()  # 设计概念
    business = scrapy.Field()  # 主营业务
    date_listing = scrapy.Field()  # 上市日期


class NewsItem(scrapy.Item):
    """
    股票的历史新闻，因为爬取到的每日新闻数量不固定，计划使用mysql数据库以字符串形式存储
    """
    code = scrapy.Field()  # 代码，唯一标识符，表示数据属于哪个股票
    # 来源 = scrapy.Field()
    标题 = scrapy.Field()
    date = scrapy.Field()
    url = scrapy.Field()
    内容 = scrapy.Field()



"""
股票功能的入口文件
"""
import os

import yangke.stock.globalVar as gv  # 所有的全局变量定义在该模块中，使用全局变量需要引入该模块
from yangke.stock.dataset.tushareData import StockData
import io
from yangke.common.config import *
from yangke.common.mysql import *
from yangke.common.crawler import start_scrapy_start

from yangke.stock.db_tables_schema import holiday_table, symbols_table


def config_program():
    """
    配置项目相关的文件夹路径、项目架构等信息
    这里定义全局变量
    :return:
    """
    # 设置标准输出流数据格式为utf-8，否则在java中调用输出的中文是乱码
    try:
        sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')  # 不同系统甚至不同开发环境中sys.stdout都可能不同
    except AttributeError:  # IDLE中，sys.stdout对象没有buffer属性
        logger.info("sys.stdout没有buffer属性，IDLE的默认输出会出现该问题")
    # -----------------初始化所有全局变量-----------------
    res = connect_mysql(user='root', passwd='111111', db='stocks', return_type="engine", use_pymysql=False)
    # res = connect_mysql(host="rm-bp1bjlvw6u3pmo742bo.mysql.rds.aliyuncs.com", user="yangke007",
    #                     passwd="YangKe08", db='stocks', return_type="engine")
    if res is None:
        logger.warning('数据库连接失败')
        gv.storage = "file"
    else:
        gv.mysql_engine = res

    # 处理命令行参数
    # 调用示例：python sis_io.py -c debug --symbol 600006 --debug
    args_define = {"kv": [
        {"short": 'c', "long": "command", "default": "debug"},
        {"short": 's', "long": "symbol", "default": "600006"},

    ], 'k': [
        {'short': 'd', 'long': 'debug', 'default': True, 'description': 'run in debug mode'}
    ]
    }
    args = get_args(args_name=args_define)
    if args is None:
        print("未传入配置参数，使用默认参数'-s 600007 --debug'")
        gv.command, gv.symbol, gv.debugMode = 'debug', '600006', True
    else:
        try:
            gv.command, gv.symbol, gv.debugMode = args.command, args.symbol, args.debug
        except AttributeError:
            print(f"传入参数{args}错误，使用默认参数'-s 600007 --debug'")
            gv.command, gv.symbol, gv.debugMode = 'debug', '600006', True
    if gv.debugMode:
        gv.logger = initLogger(logging.DEBUG)
        holdLoggingLevel(logging.DEBUG, outer=False)  # paddle会覆盖configLogging(logging.DEBUG)的设置，用该方法顶掉paddle的日志设置
    else:
        gv.logger = initLogger(logging.WARNING)
        holdLoggingLevel(logging.WARNING, outer=False)  # paddle会覆盖configLogging(logging.WARNING)的设置

    # gv.dataDirectory = os.path.split(os.path.realpath(__file__))[0]  # 这个是该文件所在的路径
    # gv.dataDirectory = os.path.dirname(gv.dataDirectory)
    gv.dataDirectory = os.getcwd()  # 使用项目所在的目录作为数据储存目录，以便可以在项目删除时方便的删除数据文件
    gv.dataDirectory = os.path.join(gv.dataDirectory, "StockData")
    if not os.path.exists(gv.dataDirectory):
        os.mkdir(gv.dataDirectory)
    if gv.storage == "file":
        gv.tsd = StockData(gv.dataDirectory)
        gv.logger.debug("股票数据存储目录：" + gv.dataDirectory)
    else:
        # 初始化假日数据表
        if not has_table("holiday"):
            # 如果节假日日历数据表不存在，则创建
            create_table(cursor=gv.mysql_engine, metadata=holiday_table)
        # 初始化符号数据表
        if not has_table("symbols"):
            create_table(cursor=gv.mysql_engine, metadata=symbols_table)

        # 股票交易数据和新闻的表无法提前初始化，因为必须知道股票代码才能创建对应的表
        gv.tsd = StockData(gv.dataDirectory, mysql_engine=res)
        gv.logger.debug("股票数据存储于mysql数据库")
    # -----------------初始化所有全局变量-----------------

    # 判断相关库是否安装
    # if existModule("tushare"):
    #     import tushare as ts
    #     gv.logger.debug("检测到tushare {}已经安装！".format(ts.__version__))
    # if existModule("tensorflow"):
    #     import tensorflow as tf
    #     gv.logger.debug("检测到tensorflow {}已经安装！".format(tf.__version__))
    # if existModule("paddle"):
    #     import paddle
    #     gv.logger.debug("检测到paddle {}已经安装！".format(paddle.__version__))
    # if existModule("torch"):
    #     import torch
    #     gv.logger.debug("检测到pytorch {}已经安装！".format(torch.__version__))


def start():
    # 程序全局设置
    config_program()

    start_time = datetime.datetime.now()  # 用来统计执行时间

    if gv.command == "download":
        gv.tsd.download(gv.symbol, append=True)

    elif gv.command == "getName":
        gv.tsd.get_name()

    elif gv.command == "getOpenLimitUp":
        gv.tsd.get_open_limit('up')

    elif gv.command == "getOpenLimitDown":
        gv.tsd.get_open_limit_and_lastdays('down')

    elif gv.command == "getSymbols":
        symbols = gv.tsd.get_all_stock_symbol()  # 会生成allsymbols.csv
        # for symbol in symbols:#传递参数太多，不适合outputStream传递，使用硬盘文件传递
        #     print("py to java:{}".format(symbol))
        # print(np.array(symbols))

    elif gv.command == "getStockNum":
        symbols = gv.tsd.get_all_stock_symbol()
        print("共有 {} 只股票".format(len(symbols)))

    elif gv.command == "getNewStocks":
        # 查询当前所有正常上市交易的股票列表 ts_code     symbol     name     area industry    list_date(上市日期)
        data = gv.tsd.api.query('stock_basic', exchange='', list_status='L',
                                fields='ts_code,symbol,name,area,industry,list_date')
        sortedData = data.sort_values(by="list_date", ascending=False)
        for index, row in sortedData[0:20].iterrows():
            print(row["list_date"], row["name"], row["area"], row["industry"], row["symbol"])

    elif gv.command == "getBottomOut":
        symbols, _ = gv.tsd.get_bottom_out(int(gv.symbol))
        print("{}天内触底反弹股票列表如下：".format(gv.symbol))
        print(symbols)

    elif gv.command == "debug":
        gv.tsd.download(gv.symbol, append=True)
        # tsd.download_daily(date.today())
        # gv.tsd.get_open_limit()
        # gv.tsd.get_bottom_out(10)
        # logging.debug("最后一个交易日：{}".format(tsd.get_last_working_day()))
        # logging.debug("查询股票{}今天的交易数据：".format(symbol, tsd.get_data_of_day(symbol)))
        # gv.logger.debug("目前上市的股票代码：\n{}".format(pd.Series(gv.tsd.get_all_stock_symbol())))
        # gv.logger.debug("最近的工作日列表:\n{}".format(pd.Series(gv.tsd.get_last_n_working_day(30))))
        # tsd.get_open_limit_of_n_trading_days(30, 'up', 'ascend')
        # gv.tsd.get_open_limit_of_n_trading_days(30, 'down', 'ascend')
        # gv.tsd.get_open_limit_and_lastdays(last_date=datetime.date.fromisoformat("2019-12-31"))
        # gv.tsd.recent_30_open_limit_down_to_file(days=30)

        # 模拟交易
        symbols = gv.tsd.get_all_stock_symbol()
        # db.test1()
        # Main.test1()

        # 神经网络预测
        # pad.test1()
        # pad.test2()
        # pad.test3()
        # pad.test4()
        # datafile = os.path.join(gv.dataDirectory, gv.symbol + '.csv')
        # pps.prediction1(datafile)

        # import prediction.pytorch_pred as ptp
        # ptp.start_mysql_service()

        import yangke.stock.prediction.pytorch_pred_stock as ptps

        ptps.lstm_stock_use_in_build_rnn_fit(gv.symbol)
        # ptps.lstm_stock_classify(gv.symbol)

        # start_mysql_service()

        start_scrapy_start(r"start_stock10jqka.py")
        # start_scrapy_spider(r"jqka_spider.py")  # 因为在start_stock10jqka.py中还初始化了数据库，因此只能从start_...开始执行爬虫

        # stocks = gv.tsd.download_all_stocks(exclude=['000029', '600145', '688086', '601696'])  # 排除掉两个停牌了很多年的傻逼股票
        # pred_df = ptps.prediction(stocks, 'fit')
        # pred_df.sort_values(by=['p_change'], inplace=True, ascending=False)
        # gv.logger.info("预测涨幅最大的5只股票为：\n{}".format(pred_df[:5]))

    else:
        gv.logger.error("Unknown command: {}".format(gv.command))
        sys.exit(0)
    end_time = datetime.datetime.now()
    print("执行用时：{}s".format((end_time - start_time)))


if __name__ == "__main__":
    start()



# import os.path
#
# from yangke.performance.tools.natural_gas import NaturalGas, logger
#
#
# def gas_properties(compositions=None, visible=False):
#     """
#     计算天然气物性。
#
#
#     :param compositions:
#     :param visible: 本函数使用excel计算天然气物性，该参数表示计算过程是否显示Excel界面，默认不显示
#     :return:
#     """
#     if compositions is None:
#         compositions = {"CH4": 0.9, "N2": 0.1}
#     k_idx = {
#         "CH4": 1,
#         "N2": 52,
#         "CO2": 54,
#         "C2H6": 2,
#         "C3H8": 3,
#         "H2O": 42,
#         "H2S": 43,
#         "H2": 41,
#         "CO": 46,
#         "O2": 53,
#         "C4H10": 4,
#         "C4H10-1": 5,
#         "C5H12": 6,
#         "C5H12-1": 7,
#         "C6H14": 9,
#         "C7H16": 14,
#         "C8H18": 15,
#         "C9H20": 16,
#         "C10H22": 17,
#         "He": 49,
#         "Ar": 51,
#         "C2H4": 18,
#         "C3H6": 19,
#         "C4H8": 20,
#     }
#
#     app = xw.App(visible=visible, add_book=False)
#     app.display_alerts = False
#     app.screen_updating = visible
#     wb = app.books.open(os.path.join(os.path.dirname(__file__), "resource/ISO6976.xlsx"))
#     sht = wb.sheets['Summary']
#     for i in range(9, 69, 1):
#         sht.range(f"C{i}").value = 0
#
#     for k, v in compositions.items():
#         idx = k_idx.get(k)
#         row = idx + 8
#         sht.range(f"C")
#
#     app.quit()
#
#
# if __name__ == "__main__":
#     comp = {
#         "CH4": 95,
#         "N2": 1,
#         "CO2": 1,
#         "C2H6": 3
#     }
#     gas_properties(comp)



import sys
import os
from PyQt5.QtWidgets import QMainWindow, QHBoxLayout, QPushButton, QComboBox, QLineEdit, QTableWidgetItem, \
    QLabel, QCheckBox, QVBoxLayout, QWidget, QApplication, QAction, QDesktopWidget, QTableWidget, qApp, \
    QDialog, QInputDialog, QMessageBox, \
    QTextEdit
from PyQt5.QtGui import QIntValidator, QRegExpValidator, QKeyEvent, QBrush, QColor, QFont
import PyQt5.QtCore as QtCore
from yangke.core import runAsAdmin
from yangke.common.win.win_x64 import find_window, get_pid_by_hwnd
from yangke.common.mysql import connect_available
import pandas as pd
from gettext import gettext as _
import yaml

from yangke.common.config import logger
from yangke.base import get_settings, is_number


class YkDialog(QDialog):
    """
    对话框类，用于在其他界面中弹出自定义窗口，且窗口可以与父窗口通信
    """
    # 以下消息的传递参数时object，其包括(type, value)，其中type是消息想要传递的参数类型，value是想传递的值
    signal_close = QtCore.pyqtSignal(object)  # 用于对话框关闭时发送关闭消息，object=(dict:value)是YkDialog关闭时的对象的值
    signal_button_clicked = QtCore.pyqtSignal(object)  # object=(QPushButton, 按钮)，YkDialog对象会自动通过sender传递

    def __init__(self, title=_("对话框"), ui_file=None, values=None):
        """
        ui_file示例
----------------------------------------------------------------------------------------------------
size:
  width:
    - 80  # <label>的宽度
    - 340  # <textField>的宽度
    - 10  # <unit>的宽度

mysql:
  inputArea:
    - # 第一个设置项
      label: "host"
      value: "localhost"
    - # 第二个设置项
      label: "port"
      value: 3306
    - # 第三个设置项
      label: "user"
      value: "root"
    - # 第三个设置项
      label: "password"
      value: "111111"
  buttons:
    - # 第一个按钮
      text: "测试连接"
      on_click: "btn_clicked"  # 按钮点击触发的事件
    - text: "Apply"
      on_click: "btn_clicked"  # 按钮点击触发的事件
----------------------------------------------------------------------------------------------------

        :param title:
        :param ui_file:
        :param values:
        """
        super(YkDialog, self).__init__()
        self.setWindowTitle(title)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        if ui_file is not None:
            self.widget = YkInputPanel(parent=self, from_file=ui_file, domain="mysql")
        self.result_panel = QLabel()
        layout = QVBoxLayout()
        layout.addWidget(self.widget)
        layout.addWidget(self.result_panel)
        self.setLayout(layout)
        if values is not None:
            self.set_values(values)

    def set_values(self, values):
        self.widget.set_values(values)

    def btn_clicked(self):
        sender = self.sender()
        self.signal_button_clicked.emit((QPushButton, sender))

    def get_value(self):
        result = self.widget.get_values_and_units(need_dict=True, need_unit=False)
        return result

    def set_result_panel(self, info):
        """
        对话框按钮点击后，可能需要显示按钮事件处理后的消息，这里显示到面板上

        :param info:
        :return:
        """
        self.result_panel.setText(info)

    def closeEvent(self, e):
        value = self.get_value()
        self.signal_close.emit((dict, value))
        self.close()


class YkWindow(QMainWindow):
    button_clicked_signal = QtCore.pyqtSignal(object)

    def __init__(self, setting_file="default"):
        # noinspection PyArgumentList
        super(YkWindow, self).__init__()
        self.setting_file = setting_file
        self.toolbar = None
        self.init_ui()
        self.init_ui1()
        self.yk_info = {}  # 用于传递临时变量的变量

    def init_ui1(self):
        self.center()
        size = get_settings("mainframe.geometry", self.setting_file)
        if len(size) > 0:
            if size == "maximize":
                self.showMaximized()
            elif size == "fullscreen":
                self.showFullScreen()
            else:
                size = eval(size)
                x, y, w, h = size
                self.setGeometry(x, y, w, h)
        self.statusBar().showMessage(_('Ready'))
        logger.debug("窗口初始化完成")
        self.show()

    def init_ui(self):
        self.setToolTip('This is a <b>QWidget</b> widget')
        btn = QPushButton('Quit', self)
        btn.setToolTip('This is a <b>QPushButton</b> widget')

        btn.resize(btn.sizeHint())
        btn.move(50, 100)
        btn.clicked.connect(self.button_clicked)

        exit_action = QAction(_('Exit'), self)
        exit_action.setShortcut('Ctrl+Q')
        exit_action.setStatusTip(_('Exit application'))
        exit_action.triggered.connect(qApp.quit)

        menu_bar = self.menuBar()
        file_menu = menu_bar.addMenu(_('File'))
        file_menu.addAction(exit_action)

        self.toolbar = self.addToolBar(_('Exit'))
        self.toolbar.addAction(exit_action)

        self.setWindowTitle(_("yk demo"))
        # self.setWindowIcon(QIcon('web.png'))

    def button_clicked(self, anything=None, anything2=None, **kwargs):
        """
        处理点击事件

        :param anything: 触发该方法的信号传入的数据，可以是任何类型
        :param anything2: 接收点击事件传入的任何参数
        :return:
        """
        sender = self.sender()  # 发送事件的组件，可能是button、YkDialog等任何拥有signal的类
        if isinstance(sender, QPushButton):
            self.statusBar().showMessage(sender.text() + ' was pressed')
        elif isinstance(sender, YkDialog):  # 说明点击事件来自自定义的子对话框
            logger.debug(_("接收到自定义窗口信号！"))
            if sender.windowTitle() == "MySQL连接设置":  # 判断消息具体来自哪一个对话框实例
                if anything[0] == QPushButton:  # 说明是按钮点击事件
                    if anything[1].text() == _("测试连接"):
                        mysql_settings = sender.get_value()
                        if connect_available(**mysql_settings):
                            sender.set_result_panel("连接成功")
                        else:
                            sender.set_result_panel("连接失败")
                    elif anything[1].text() == _("应用"):
                        mysql_info = sender.get_value()
                        self.yk_info.update(mysql_info)
                        if self.setting_file is not None:
                            settings = get_settings(setting_file=self.setting_file)
                            mysql_ = settings.get("mysql") or {}
                            mysql_.update(mysql_info)
                            # settings.update({"mysql": mysql_})  # mysql_是浅拷贝，无需再次更新settings
                            with open(self.setting_file, "w") as f:
                                yaml.dump(settings, f)
                if anything[0] == dict:  # 说明是子对话框关闭时间，YkDialog关闭时会将对话框中收集的数据发送出来
                    # self.set_mysql(anything[1])
                    pass  # 关闭对话框时无需执行额外操作

    def center(self):
        qr = self.frameGeometry()  # 获得窗口
        cp = QDesktopWidget().availableGeometry().center()  # 获得屏幕中心点
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def closeEvent(self, event):
        #     reply = QMessageBox.question(self, "Message", "Are you sure to quit?",
        #                                  QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        #     if reply == QMessageBox.Yes:
        #         event.accept()
        #     else:
        #         event.ignore()
        print(_("close event triggered"))
        event.accept()

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        if a0.key() == QtCore.Qt.Key_Escape:
            self.close()

    def init_mysql(self, ui_file=None):
        """
        初始化mysql数据库，用户的更改会实时保存到settings_file中

        :param ui_file: mysql配置面板的ui文件，使用YkInputPanel(ui_file)进行面板生成，如果需要自定义面板，需要重写init_mysql方法
        :return:
        """
        mysql = get_settings("mysql", setting_file=self.setting_file) or {}
        host = mysql.get('host') or "localhost"
        port = mysql.get('port') or "3306"
        username = mysql.get('user') or "root"
        password = mysql.get('passwd') or '111111'
        self.yk_info["host"] = host
        self.yk_info["port"] = port
        self.yk_info["user"] = username
        self.yk_info["passwd"] = password

        dialog = YkDialog("MySQL连接设置", ui_file=ui_file, values=self.yk_info)
        dialog.signal_button_clicked.connect(self.button_clicked)  # 关联自定义对话框的按钮单击事件
        dialog.signal_close.connect(self.button_clicked)  # YkDialog关闭时发送yk_signal信号
        dialog.exec()

    def set_digits(self):
        digits_str, ok = QInputDialog.getText(self, "设置数据格式", "保留小数位数（最大位数）", QLineEdit.Normal, "2")
        if ok and digits_str.isnumeric():
            self.digits = int(digits_str)

    def set_font(self, font_size=10):
        font_size, ok_pressed = QInputDialog.getText(self, "设置界面字体大小", "字体大小（px）:", QLineEdit.Normal, "")
        if ok_pressed and font_size.isnumeric():
            font_size = int(font_size)
            self.setFont(QFont("Microsoft YaHei", font_size))

    def about(self, tag=None):
        """
        关于菜单的默认动作

        :return:
        """
        print(self.size())
        # QMessageBox.information(self, "关于", "Powered by open-source software", QMessageBox.Yes)
        dialog = QDialog()
        # 这个btn和上面的btn不是一个
        edit = QTextEdit()
        edit.setHtml('''
            <h1>lib4python<h1>
            <h2>yangke 1.6.3<h2>        
            <p>开发于PyCharm 2020.3.2 (Community Edition)<p>
            <p>构建与2021年03月05日<p>
            Runtime version: 11.0.9.1+11-b1145.63 amd64
            VM: OpenJDK 64-Bit Server VM by JetBrains s.r.o.
            Windows 10 10.0
            <p>Powered by open-source software<p>
        ''')
        # edit.setReadOnly(True)
        layout = QVBoxLayout(dialog)
        # dialog.setLayout()
        layout.addWidget(edit)

        dialog.setWindowTitle('关于')
        # 让对话框以模式状态显示，即显示时QMainWindow里所有的控件都不可用，除非把dialog关闭
        dialog.setWindowModality(QtCore.Qt.ApplicationModal)
        dialog.resize(900, 300)
        dialog.exec()

    def help(self, tag=None):
        """
        帮助按钮的默认动作

        :return:
        """
        QMessageBox.information(self, "帮助", "暂不支持帮助", QMessageBox.Yes)


class YkItem(QWidget):
    def __init__(self, label="Label",
                 value=None, value_selected=None, value_validator=None,
                 unit=None, unit_selected=None,
                 size=None, **kwargs):
        """
        一种输入项，可以带下拉列表单位

        :param label:
        :param value:
        :param value_selected:
        :param unit:
        :param unit_selected:
        :param type: checkbox/None/textLine  -> kwargs
        :param checked: 复选框类型中，被选中的标签  -> kwargs
        :param labels: 复选框类型中，可选项的标签列表  -> kwargs
        """
        # noinspection PyArgumentList
        super(YkItem, self).__init__()
        self.type = kwargs.get("type")
        if self.type is None:
            self.label = _(label)
            self.value = None
            if value is not None:
                if isinstance(value, list):
                    self.value = QComboBox(self)
                    self.value.addItems(value)
                    self.value.setCurrentText(value_selected)
                else:
                    self.value = QLineEdit(str(value))
                    if value_validator is not None:
                        validator_types = value_validator.split(",")
                        if validator_types[0] == "int":
                            validator = QIntValidator(self.value)
                            validator.setRange(int(validator_types[1]), int(validator_types[2]))
                            self.value.setValidator(validator)
                            self.value.setPlaceholderText(value_validator)
                        elif validator_types[1] == "regexp":
                            validator = QRegExpValidator(self)
                            # todo 正则表达式设置QLineEdit的允许输入值
                            self.value.setValidator(validator)
            self.unit = None
            if unit is not None:
                if isinstance(unit, list):
                    self.unit = QComboBox(self)
                    self.unit.addItems(unit)
                    self.unit.setCurrentText(unit_selected)
            self.size = size
            self.init_ui()
        elif self.type == "checkbox":
            items = kwargs.get("items")
            checked = kwargs.get("checked")
            self.checkbox_layout = QHBoxLayout()

            self.label = label
            self.checkbox_layout.addWidget(QLabel(self.label))
            for item in items:
                checkbox = QCheckBox(item, self)
                if item in checked:
                    checkbox.setChecked(True)
                self.checkbox_layout.addWidget(checkbox)

            self.checkbox_layout.addStretch()
            self.value = self.checkbox_layout
            self.setLayout(self.checkbox_layout)
            # self.setStyleSheet("border:1px solid grey")

    def init_ui(self):
        # self.setPalette(QPalette(QtCore.Qt.red))
        # self.setAutoFillBackground(True)
        self.setMinimumSize(20, 5)

        h_box = QHBoxLayout()
        label = QLabel(self.label, self)
        # ----------------------------设置组件大小--------------------------------
        if self.size is not None:
            width = self.size.get("width")
            if isinstance(width, int):
                width = [width, width, width]
            if isinstance(width, list):
                if len(width) >= 1:
                    label.setFixedWidth(int(width[0]))
                if len(width) >= 2:
                    self.value.setFixedWidth(int(width[1]))
                if len(width) >= 3 and self.unit is not None:
                    self.unit.setFixedWidth(int(width[2]))
        # ----------------------------设置组件大小--------------------------------

        # noinspection PyArgumentList
        h_box.addWidget(label)
        # noinspection PyArgumentList
        h_box.addWidget(self.value)
        if self.unit is not None:
            h_box.addWidget(self.unit)
        # h_box.setSpacing(0)
        h_box.addStretch()
        self.setLayout(h_box)
        # self.setStyleSheet("background-color:green;padding:2;")

    def get_value(self):
        value = None
        if self.type is None or self.type.lower() == "textline":
            if isinstance(self.value, QLineEdit):
                value = self.value.text()
            elif isinstance(self.value, QComboBox):
                value = self.value.currentText()
            return value
        elif self.type.lower() == "checkbox":
            value = []
            for item in self.findChildren(QCheckBox):
                item: QCheckBox = item
                if item.isChecked():
                    value.append(item.text())
            return value

    def get_unit(self):
        unit = None
        if self.type is None or self.type.lower() == "textline":
            if self.unit is not None:
                unit = self.unit.currentText()
        return unit

    def get_value_and_unit(self):
        return self.get_value(), self.get_unit()

    def set_value(self, value):
        value = str(value)
        if isinstance(self.value, QLineEdit):
            self.value.setText(value)
        elif isinstance(self.value, QComboBox):
            self.value.setCurrentText(value)
        elif isinstance(self.value, QHBoxLayout):  # 说明是复选框
            for item in self.findChildren(QCheckBox):
                item: QCheckBox = item
                if item.text() in value:
                    item.setChecked(True)

    def set_unit(self, unit):
        if unit is not None:
            self.unit.currentText(unit)


# @auto_save_para  # 使用硬盘变量，即初始化为None的成员变量会以硬盘方式存储，任何改变都会直接写入硬盘
class InjectPanel(QWidget):
    def __init__(self):
        """
        初始化函数
        """
        # noinspection PyArgumentList
        super(InjectPanel, self).__init__()
        self.resize(800, 600)
        layout = QVBoxLayout()
        layout.setSpacing(10)
        layout.setAlignment(QtCore.Qt.AlignTop)
        # 硬盘变量行不要出现非注释用的“#”字符和多个”#“字符，否则可能会导致auto_save_para装饰器出错
        self.title = None  # 该变量会保存在硬盘上
        self.pid = None  # 初始化为0不是硬盘变量
        self.dll_path = None

        # 添加自定义部件（MyWidget）
        self.title_item = YkItem("目标窗口", self.title)
        self.pid_item = YkItem("目标进程", self.pid)
        self.dll_item = YkItem("待注入dll", self.dll_path)
        inject_btn = QPushButton("注入")
        inject_btn1 = QPushButton("注入")
        inject_btn2 = QPushButton("注入")

        h_box = QHBoxLayout()
        h_box.addStretch()
        # noinspection PyArgumentList
        h_box.addWidget(inject_btn)
        # noinspection PyArgumentList
        h_box.addWidget(inject_btn1)
        # noinspection PyArgumentList
        h_box.addWidget(inject_btn2)

        # 放入布局内
        # noinspection PyArgumentList
        layout.addWidget(self.title_item)
        # noinspection PyArgumentList
        layout.addWidget(self.pid_item)
        # noinspection PyArgumentList
        layout.addWidget(self.dll_item)
        layout.addStretch()
        layout.addLayout(h_box)

        self.setLayout(layout)
        # self.setStyleSheet("margin:2;padding:2")
        self.setWindowTitle("zr")
        self.show()
        inject_btn.clicked.connect(self.btn_clicked)

    def btn_clicked(self):
        sender = self.sender()
        if sender.text() == "注入":
            logger.debug("注入点击")
            zr_path = os.path.join(os.path.dirname(__file__), "zr.exe")
            if not os.path.exists(zr_path):
                logger.critical(f"zr.exe not found in {zr_path}")
            self.title = self.title_item.get_text_edit().text()
            windows_dict = find_window(self.title, exact=False)
            hwnds = [hwnd1 for hwnd1 in windows_dict.keys()]
            if len(windows_dict) == 0:
                logger.info(f"未找到窗口：{self.title}")
                return
            elif len(windows_dict) > 1:
                logger.warning(f"找到多个符合条件的窗口:{''.join(str(hwnds))}")

            hwnd = hwnds[0]
            title = windows_dict.get(hwnd)
            _, _pid = get_pid_by_hwnd(hwnd)
            self.pid = _pid
            self.dll_path = self.dll_item.get_text_edit().text()
            cmd = f'"{zr_path}" -pid {self.pid} -dll "{self.dll_path}"'
            runAsAdmin(cmd)


def layout_to_widget(layout):
    """
    讲PyQt5中的layout转换为Widget。
    用于：
    在QSplitter中添加内容时，只能使用QWidget类对象，如果是用户创建的QVBoxLayout内容，则无法添加，可以使用该方法转换后添加。

    :param layout:
    :return:
    """

    class YKWidget(QWidget):
        def __init__(self):
            super(YKWidget, self).__init__()
            self.setLayout(layout)

    widget = YKWidget()
    return widget


def QYKFigure(x=None, y=None, xlim=None, ylim=None,
              fig_type=None, figsize=None, dpi=None, facecolor=None, edgecolor=None,
              linewidth=0.0,
              frameon=None, subplotpars=None, tight_layout=None, constrained_layout=None,
              title=None, xlabel=None):
    """
    在PyQt5面板中绘制matplotlib图形，该方法返回一个QWidget图形对象

    :param ylim: (bottom: float, top: float)
    :param xlim: (bottom: float, top: float)
    :param xlabel:
    :param x:
    :param y:
    :param fig_type: scatter/bar/hist/curve
    :param figsize:
    :param dpi:
    :param facecolor:
    :param edgecolor:
    :param linewidth:
    :param frameon:
    :param subplotpars:
    :param tight_layout:
    :param constrained_layout:
    :param title:
    :return:
    """
    import matplotlib
    matplotlib.use("Qt5Agg")
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.figure import Figure
    from matplotlib.font_manager import _rebuild
    import matplotlib.pyplot as plt
    from matplotlib.axes._subplots import Axes

    plt.rcParams['font.sans-serif'] = ['SimHei']
    plt.rcParams['axes.unicode_minus'] = False
    import numpy as np

    _rebuild()

    class Qt5Figure(FigureCanvas):
        def __init__(self, figsize=None, dpi=None, facecolor=None, edgecolor=None, linewidth=0.0,
                     frameon=None, subplotpars=None, tight_layout=None, constrained_layout=None):
            self.figure = Figure(figsize=figsize, dpi=dpi, facecolor=facecolor, edgecolor=edgecolor,
                                 linewidth=linewidth, frameon=frameon, tight_layout=tight_layout,
                                 subplotpars=subplotpars, constrained_layout=constrained_layout)

            # 在父类中激活Figure窗口，此句必不可少，否则不能显示图形
            super(Qt5Figure, self).__init__(self.figure)
            self.axes: Axes = self.figure.add_subplot(111)
            if title is not None:
                self.figure.suptitle(title)
            self.axes.set_xlabel(xlabel)
            self.axes.set_ylim(auto=True)
            self.axes.set_xlim(auto=True)
            self.axes.grid(axis="both", color="y", linestyle=":", linewidth=1)

            if fig_type is not None:
                if fig_type == "scatter":
                    self.scatter(x, y)
                elif fig_type == "bar":
                    self.bar(x, y)
                elif fig_type == "hist":
                    self.hist(x)
                elif fig_type == "curve":
                    self.curve(x, y)
            else:
                self.curve(x, y)
            # self.figure.tight_layout()

        def add_fig(self, fig_type, x=None, y=None, **kwargs):
            """
            在当前图中添加图。
            一般用于同一个坐标系多条曲线的情况，添加第二条曲线时即可使用该方法。

            :param fig_type: 新图的类型，curve/scatter/bar/hist
            :param x:
            :param y:
            :return:
            """
            if fig_type == "scatter":
                self.scatter(x, y, **kwargs)
            elif fig_type == "bar":
                self.bar(x, y, **kwargs)
            elif fig_type == "hist":
                self.hist(x, **kwargs)
            elif fig_type == "curve":
                self.curve(x, y, **kwargs)
            self.figure.tight_layout()

        def curve(self, x, y, **kwargs):
            if isinstance(x, float) or isinstance(x, int):
                x = [x]
            if isinstance(y, float) or isinstance(y, int):
                y = [y]
            self.axes.plot(x, y, **kwargs)

        def scatter(self, x, y, **kwargs):
            """
            散点图
            :param x:
            :param y:
            :return:
            """
            if isinstance(x, float) or isinstance(x, int):
                x = [x]
            if isinstance(y, float) or isinstance(y, int):
                y = [y]
            x = np.array(x)
            y = np.array(y)
            if x.shape != y.shape:
                if x.shape[0] == y.shape[0]:
                    x = np.repeat(x, y.shape[1])
                elif x.shape[0] == y.shape[1]:
                    x = np.tile(x, y.shape[0])
                y = y.flatten()
            if x.shape != y.shape:
                raise ValueError("散点图的数据列表长度不等，且无法扩展到相同维度")
            self.axes.scatter(x, y, **kwargs)

        def bar(self, x, y, **kwargs):
            """
            柱状图
            :param x: 列表或数值
            :param y: 列表或数值
            :return:
            """
            if isinstance(x, float) or isinstance(x, int):
                x = [x]
            if isinstance(y, float) or isinstance(y, int):
                y = [y]
            self.axes.bar(x, y, **kwargs)

        def hist(self, value, bins=10):
            """
            直方图
            :param value:
            :param bins:
            :return:
            """
            self.axes.hist(value, 10)

        def show_control_panel(self):
            """
            显示图片的控制面板

            :return:
            """
            # todo
            pass

        def hide_control_panel(self):
            """
            隐藏控制面板

            :return:
            """
            # todo
            pass

    return Qt5Figure()


def set_menu_bar(window: QMainWindow, from_file="ui_menu.yaml"):
    """
    根据ui_menu.yaml文件定义的菜单栏菜单项信息设置window窗口的菜单栏

    ui_menu.yaml示例文件
------------------------------------------------------------------------------------
menu:
  - # 第一个菜单ribbon
    name: "文件"
    items:
      - action_name: "设置数据文件"
        short_cut: "Ctrl+Q"
        connect: "self.set_data_file"
  - # 种群图类型
    name: "设置"
    items:
      - action_name: "计算设置"
        connect: "self.set_calculate"
      - action_name: "遗传算法设置"
        connect: "self.set_algorithm"
      - action_name: "图像显示设置"
        connect: "self.set_display"
  -
    name: "帮助"
    items:
      - action_name: "关于"
      - action_name: "帮助"
------------------------------------------------------------------------------------

    :param window:
    :param from_file:
    :return:
    """
    menu_bar = window.menuBar()
    menus = get_settings("menu", setting_file=from_file)
    for ribbon in menus:
        temp_ribbon = menu_bar.addMenu(_(ribbon.get("name")))
        actions = ribbon.get("items")
        for item in actions:
            temp_action = QAction(item.get("action_name"), window)
            if item.get("short_cut") is not None:
                temp_action.setShortcut(item.get("short_cut"))
            temp_connect: str = item.get("connect")
            if temp_connect is not None:
                if temp_connect.startswith("self."):
                    temp_connect = temp_connect.replace("self.", "")
                temp_action.triggered.connect(eval(f"window.{temp_connect}"))
            temp_ribbon.addAction(temp_action)


logger.debug("加载YkDataTableWidget")


class YkDataTableWidget(QTableWidget):
    """
    数据表格组件，配置的table_data.yaml文件参见ui/table_data1.yaml;ui/table_data2.yaml;table_data_single_calculate.yaml

以下yaml文件内容向表格中添加一个按钮，按钮链接表格组件的self.single_calculate方法
---------------------- table.yaml ----------------------
button:
  - name: "计算"
    range: "(2,4)"  # 按钮位于表格的第三列四行
    connect: "self.single_calculate"
--------------------------------------------------------
    self.single_calculate方法必须先传入，然后初始化，例如：
---------------------- example.py ----------------------
def calculate:
    # calculate中需要完成的操作
    pass

YkDataTableWidget.single_calculate = calculate
table_widget = YkDataTableWidget(from_file = "table.yaml")
--------------------------------------------------------
    """

    def __init__(self, from_file="table_data.yaml", **kwargs):
        super(YkDataTableWidget, self).__init__()
        if not os.path.exists(from_file):
            logger.warning(f"YkDataTableWidget的数据设置文件<{from_file}>未找到，请确定文件路径是否正确！")

        settings = get_settings(setting_file=from_file)
        data = settings.get("data") or {}
        buttons = settings.get("button") or {}
        width1 = settings.get("width")

        self.columns = 20
        self.rows = 100
        self.setRowCount(self.rows)
        self.setColumnCount(self.columns)
        # 初始化var_loc用于存储表格中记录的数值，格式为：
        # {"title_1": {"label_1": (x_left_top, y_left_top), "label_2": (x1, y1)}, "title_2": {}, ...}
        # 对应table_data.yaml的格式
        self.var_loc = {}
        self.buttons = {}
        for domain in data:
            title = domain.get("title")
            # -------------------- 处理range参数，生成可以直接使用的x0,y_left_top,width,rows-------------------------------
            domain_range = eval(domain.get("range")) or (0, 0)
            x_left_top, y_left_top, width, height = (0, 0, 3, 0)
            if len(domain_range) == 2:
                x_left_top, y_left_top = domain_range
            elif len(domain_range) == 3:
                x_left_top, y_left_top, width = domain_range
            elif len(domain_range) == 4:
                x_left_top, y_left_top, width, height = domain_range
            # -------------------- 处理range参数，生成可以直接使用的x0,y_left_top,width,rows-------------------------------
            # -------------------- 处理background/foreground/align参数 -------------------------------
            background = domain.get("background") or "#ffffff"
            foreground = domain.get("foreground") or "#000000"
            if background.startswith("QBrush"):
                background = eval(background)
            else:
                background = QBrush(QColor(background))
            if foreground.startswith("QBrush"):
                foreground = eval(foreground)
            else:
                foreground = QBrush(QColor(foreground))
            row_index, col_index = y_left_top, x_left_top
            if title is not None:  # 表格中的小分区
                items = domain.get("items")
                height = len(items)

                if height == 0:
                    height = len(items)
                self._set_row_column_count(x_left_top, y_left_top, width, height)
                align = self._get_align(item_dict=domain, default="AlignCenter")

                # -------------将表格第一行单元格合并，并将标题内容填入第一行，且设置第一行style----------------
                self.setSpan(row_index, col_index, 1, width)
                self.setItem(row_index, col_index, QTableWidgetItem(title))
                self.item(row_index, col_index).setBackground(
                    background)  # 这里itemAt()用来设置颜色只有第一条设置语句有效，后续无效，测试发现item()函数可用
                self.item(row_index, col_index).setForeground(foreground)
                self.item(row_index, col_index).setTextAlignment(align)
                self.var_loc[title] = {}
                # -------------将表格第一行单元格合并，并将标题内容填入第一行，且设置第一行style----------------
                # --------------------------填充数据行内容------------------------------
                for i, item in enumerate(items):
                    label = item.get("label") or ""
                    value = item.get("value") or ""
                    unit = item.get("unit") or ""
                    x = row_index + i + 1
                    merge_label = item.get("merge_label_row_col")
                    if merge_label is not None:
                        merge_row, merge_col = eval(merge_label)
                        self.setSpan(x, col_index, merge_row, merge_col)
                    merge_value = item.get("merge_value_row_col")
                    if merge_value is not None:
                        merge_row, merge_col = eval(merge_value)
                        self.setSpan(x, col_index + 1, merge_row, merge_col)
                    self.setItem(x, col_index, QTableWidgetItem(label))
                    self.setItem(x, col_index + 1, QTableWidgetItem(value))
                    self.setItem(x, col_index + 2, QTableWidgetItem(unit))
                    self.var_loc[title].update({label: (x, col_index + 1)})
                # --------------------------填充数据行内容------------------------------
                # --------------------------设置数据行格式------------------------------
                for i in range(1, height + 1):
                    x = row_index + i
                    self.item(x, col_index).setBackground(background)
                    self.item(x, col_index).setForeground(foreground)
                    self.item(x, col_index + 1).setBackground(background)
                    self.item(x, col_index + 1).setForeground(foreground)
                    self.item(x, col_index + 2).setBackground(background)
                    self.item(x, col_index + 2).setForeground(foreground)
                # --------------------------设置数据行格式------------------------------
            else:
                label = domain.get("label")
                value = domain.get("value")
                unit = domain.get("unit")
                if height == 0:
                    height = 1
                self._set_row_column_count(x_left_top, y_left_top, width, height)
                merge_label = domain.get("merge_label_row_col")
                merge_row_label, merge_col_label = 1, 1
                if merge_label is not None:
                    merge_row_label, merge_col_label = eval(merge_label)
                    self.setSpan(row_index, col_index, merge_row_label, merge_col_label)
                merge_value = domain.get("merge_value_row_col")
                merge_row_value, merge_col_value = 1, 1
                if merge_value is not None:
                    merge_row_value, merge_col_value = eval(merge_value)
                    self.setSpan(row_index, col_index + 1, merge_row_value, merge_col_value)

                align = self._get_align(domain, "AlignLeft")

                # --------------------------------- 按照参数类型添加不同组件 ---------------------------------
                if isinstance(value, bool):  # 如果参数取值为bool类型，则使用QCheckBox组件就可以很好的满足要求
                    check_box = QCheckBox(label)
                    check_box.setChecked(value)
                    self.setSpan(row_index, col_index, merge_row_label, merge_col_label)
                    self.setCellWidget(row_index, col_index, check_box)
                else:
                    self.setItem(row_index, col_index, QTableWidgetItem(label))
                    self.setItem(row_index, col_index + merge_col_label, QTableWidgetItem(value))
                    self.setItem(row_index, col_index + merge_col_label + merge_col_value, QTableWidgetItem(unit))
                # --------------------------------- 按照参数类型添加不同组件 ---------------------------------

                # --------------------------------- 按照参数类型更新参数所在的位置 ---------------------------------
                if isinstance(value, bool):  # 布尔型设置项的取值就是布尔型组件本身的坐标
                    self.var_loc.update({label: (row_index, col_index)})
                else:
                    self.item(row_index, col_index).setTextAlignment(align)
                    self.item(row_index, col_index).setBackground(background)
                    self.item(row_index, col_index).setForeground(foreground)
                    self.item(row_index, col_index + merge_col_label).setBackground(background)
                    self.item(row_index, col_index + merge_col_label).setForeground(foreground)
                    self.item(row_index, col_index + merge_col_label + merge_col_value).setBackground(background)
                    self.item(row_index, col_index + merge_col_label + merge_col_value).setForeground(foreground)
                    self.var_loc.update({label: (row_index, col_index + merge_col_label)})
                # --------------------------------- 按照参数类型更新参数所在的位置 ---------------------------------
        for button in buttons:
            name = button.get("name")
            q_btn = QPushButton(name)
            col_index, row_index = eval(button.get("range"))
            self.setCellWidget(row_index, col_index, q_btn)
            if button.get("connect") is not None:
                func = eval(button.get("connect"))
                q_btn.clicked.connect(func)

            self.buttons.update({name: q_btn})

        if width1 is not None:
            width1 = eval(width1)
            for i, w in enumerate(width1):
                self.setColumnWidth(i, w)

    def _set_row_column_count(self, x, y, width, height):
        """
        根据小区域的大小和位置设置表格宽高，保证表格可以容纳整个小区域

        :param x:
        :param y:
        :param height:
        :param width:
        :return:
        """
        # 设置表格长宽
        if width + x > self.columns:
            self.columns = width + x
            self.setColumnCount(self.columns)
        if height + y > self.rows:
            self.rows = height + y
            self.setRowCount(self.rows)

    @staticmethod
    def _get_align(item_dict, default="AlignCenter"):
        """
        从当前字典中获取align值，如果不存在，则生成default对应的align值，返回QtCore.Qt.AlginCenter等对象

        :param item_dict:
        :param default:
        :return:
        """
        align = item_dict.get("align") or "AlignCenter"
        if align.startswith("Qt"):
            align = eval(f"QtCore.{align}")
        elif align.startswith("QtCore"):
            align = eval(align)
        else:
            align = eval(f"QtCore.Qt.{align}")
        return align

    def get_var_location(self, var_name: str):
        """
        获取某个变量在表格中的位置
        :param var_name: 参数名称，如果表格时按照小区域分割的，则参数名称以 domain.label 的格式传入
        :return: tuple(x,y)
        """
        var_name = var_name.split(".")
        temp = self.var_loc
        for lvl in var_name:
            temp = temp.get(lvl)
            if isinstance(temp, tuple):
                return temp
        print("所查找的变量不存在")
        return -1, -1

    def set_value(self, var_name: str, value=""):
        """
        设置表格中某个参数的值
        :param var_name: 参数名称，如果表格时按照小区域分割的，则参数名称以 domain.label 的格式传入
        :param value: 需要设置的值
        :return:
        """
        x, y = self.get_var_location(var_name)
        if x != -1:  # 说明存在制定字符串对应的变量
            if type(value) == bool:  # 布尔型变量设置checked状态
                self.cellWidget(x, y).setChecked(value)
            else:
                value = str(value)
                self.item(x, y).setText(value)  # 不能使用setItem()方法，否则会改变单元格样式
        else:
            logger.debug(f"{var_name}未找到")

    def get_value(self, var_name: str):
        """
        获取表格中某个参数的值
        :param var_name: 参数名称，如果表格时按照小区域分割的，则参数名称以 domain.label 的格式传入
        :return:
        """
        x, y = self.get_var_location(var_name)
        try:
            # 常规的文本类型直接返回text即可
            result = self.item(x, y).text()
        except AttributeError:  # 可能是QCheckBox类型，需要返回是否选中的bool值
            result = self.cellWidget(x, y)
            result = result.isChecked()
        return result

    def get_button(self, name):
        """
        根据按钮的文字获取按钮对象
        :param name:
        :return:
        """
        q_btn = self.buttons.get(name)
        return q_btn

    def display_dataframe(self, df: pd.DataFrame, row_index: int = 0, col_index: int = 0,
                          index="", header="", digits=None):
        """
        将pandas的DataFrame数据显示到YkDataTableWidget上

        :param df:
        :param df: dataframe数据
        :param row_index: 显示区域左上角的行索引
        :param col_index: 显示区域左上角的列索引
        :param index: 是否写入df的行标题，默认是写入的。如果是None，则不写入。
        :param header: 是否写入df的列标题，默认是写入原标题的。如果是None，则不写入。
        :param digits: 数据类型的最大小数点后显示位数
        :return:
        """
        if index is not None:
            df1 = df.reset_index()
        else:
            df1 = df
        if self.rowCount() < row_index + df1.shape[0] + 1:
            self.setRowCount(row_index + df1.shape[0] + 20)
        if self.columnCount() < col_index + df1.shape[1] + 1:
            self.setColumnCount(row_index + df1.shape[0] + 20)
        values = df1.values

        if header is not None:
            for j, col_name in enumerate(df1.columns):
                # setItem(row_index, col_index, QTableWidgetItem(label))
                self.setItem(row_index, j, QTableWidgetItem(str(col_name)))
            row_index = row_index + 1
        for i, row in enumerate(values):
            for j, cell in enumerate(row):
                cell = str(cell).strip()
                x = row_index + i
                y = col_index + j
                if is_number(cell) and digits is not None:
                    cell = str(round(float(cell), digits))
                self.setItem(x, y, QTableWidgetItem(cell))


def YkInputPanel(parent=None, from_file="ui_data.yaml", domain=None, return_type="widget",
                 values: list = None):
    """
    输入框面板

    # 输入界面类型一般如下：
    --------------------------------------
    |  <label1>  <textField1> <unit1>    |
    |  <label2>  <textField2> <unit2>    |
    |  <label3>  <textField3> <unit3>    |
    |                                    |
    |          <btn_apply>   <btn_ok>    |
    --------------------------------------

    实例ui_data.yaml文件
----------------------------------------------------------------
size:
  width:
    - 160  # <label>的宽度
    - 140  # <textField>的宽度
    - 80  # <unit>的宽度
algorithm:
  inputArea:
    - # 第一个设置项
      label: "种群大小"
      value: 50
    - # 第二个设置项
      label: "遗传代数"
      value: 200
  buttons:
    - # 第一个按钮
      text: "应用"
      on_click: "btn_clicked"  # 按钮点击触发的事件
    - text: "确定"
      on_click: "btn_clicked"  # 按钮点击触发的事件
----------------------------------------------------------------

    :param parent: 父组件，当有按钮时，必须传入父组件，因为按钮事件需要绑定父组件的方法
    :param values: 输入框的显示值，如果不设置，则为默认的初始值，设置的话必须一一对应每一个值，不能省略
    :param domain: ui定义文件中的选项，一个ui文件中中可以定义多个输入面板，该值表示选用哪个
    :param from_file: ui定义文件
    :param return_type: layout/widget，返回的类型，只有返回widget类型时，才拥有取值的方法
    :return:
    """
    if domain is None:  # 则将domain设置为第一个有inputArea的组件面板
        settings = get_settings(setting_file=from_file)
        for k, v in settings.items():
            if v.get("inputArea") is not None:
                domain = k
    if domain is None:
        return None
    settings = get_settings(domain, setting_file=from_file)
    inputArea = settings.get('inputArea')
    size = get_settings("size", setting_file=from_file)
    yk_items = []
    for i, item in enumerate(inputArea):
        if values is not None:
            item["value"] = values[i]  # 用调用方法传入的默认值替换ui文件中的默认值
        yk_items.append(YkItem(size=size, **item))

    btn_box = QHBoxLayout()
    btn_box.addStretch()
    for btn in settings.get('buttons'):
        btn1 = QPushButton(btn.get('text'))
        connected = f"parent.{btn.get('on_click')}"
        btn1.clicked.connect(eval(connected))
        btn_box.addWidget(btn1)

    v_box = QVBoxLayout()
    for item in yk_items:
        v_box.addWidget(item)
    v_box.addStretch()
    v_box.addItem(btn_box)

    class Panel(QWidget):
        def __init__(self):
            super(Panel, self).__init__()
            self.setLayout(v_box)
            self.name = domain
            self.values = []
            self.units = []

        def get_values_and_units(self, need_unit=True, need_dict=False):
            """
            获取YkInputPanel对象的数值和单位

            :param need_unit: 是否需要单位，不需要则只返回数值
            :param need_dict: 是否需要返回dict类型的数据，默认返回列表类型数据
            :return:
            """
            if not need_dict:
                self.values = []
                self.units = []
                for it in yk_items:
                    self.values.append(it.get_value())
                    self.units.append(it.get_unit())
                if need_unit:
                    return self.values, self.units
                else:
                    return self.values
            else:
                value_dict = {}
                unit_dict = {}
                for it in yk_items:
                    value_dict.update({it.label: it.get_value()})
                    unit_dict.update({it.label: it.get_unit()})
                if need_unit:
                    return {"value": value_dict, "unit": unit_dict}
                else:
                    return value_dict

        def set_values(self, values=None):
            """
            设置输入面板中各项的值（按顺序）

            :param values: 值的列表，或｛label:value｝的字典
            :return:
            """
            if values is None:
                return
            if isinstance(values, dict):
                for it in yk_items:
                    if it.label in values.keys():
                        it.set_value(values[it.label])
            elif isinstance(values, list):
                for i, it in enumerate(yk_items):
                    it.set_value(values[i])

    if return_type == "widget":
        return Panel()
    else:
        return v_box


if __name__ == "__main__":
    app = QApplication(sys.argv)

    w1 = InjectPanel()
    sys.exit(app.exec_())



import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate, scipy.optimize
import pandas as pd


def get_intersection_of_curve(x, y, func):
    x = np.array(x)
    y1 = np.array(y)
    y1, x = sort_an_array_by_another_array(y1, x)
    y2 = [func(x1) for x1 in x]

    interp1 = scipy.interpolate.InterpolatedUnivariateSpline(x, y1)
    interp2 = scipy.interpolate.InterpolatedUnivariateSpline(x, y2)

    def difference(x):
        return np.abs(interp1(x) - interp2(x))

    x_at_crossing = scipy.optimize.fsolve(difference, x0=3.0)

    return x_at_crossing, interp1(x_at_crossing)


def sort_an_array_by_another_array(sorted_array, by_array, ascending=True):
    """
    根据by_array的大小顺序，对两个数组都进行排序，返回两个数组排序后的结果

    :param sorted_array:
    :param by_array:
    :param ascending: 是否升序
    :return:
    """
    xy = pd.DataFrame([sorted_array, by_array])
    xy = xy.sort_values(by=1, axis=1, ascending=ascending)  # axis表示按行还是列排序，by表示指定排序的行号或列号，这里表示按行号为1的行排序
    return xy.iloc[0], xy.iloc[1]



# -*- coding: utf-8 -*-

# Scrapy settings for stock10jqka project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

BOT_NAME = 'stock10jqka'

SPIDER_MODULES = ['stock10jqka.spiders']
NEWSPIDER_MODULE = 'stock10jqka.spiders'

# Crawl responsibly by identifying yourself (and your website) on the user-agent
# USER_AGENT = 'stock10jqka (+http://www.yourdomain.com)'

# Obey robots.txt rules
ROBOTSTXT_OBEY = False

# Configure maximum concurrent requests performed by Scrapy (default: 16)
# CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See https://docs.scrapy.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
DOWNLOAD_DELAY = 1
# The download delay setting will honor only one of:
# CONCURRENT_REQUESTS_PER_DOMAIN = 16
# CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
# TELNETCONSOLE_ENABLED = False

# Override the default request headers:
DEFAULT_REQUEST_HEADERS = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
    'Accept-Language': 'en',
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.132 Safari/537.36 Edg/80.0.361.66',
}

# Enable or disable spider middlewares  中间件使用selenium进行爬取网页，因为现在大部分网页是动态的，scrapy原生组件不支持动态网页
# See https://docs.scrapy.org/en/latest/topics/spider-middleware.html
SPIDER_MIDDLEWARES = {
    'stock10jqka.middlewares.Stock10JqkaSpiderMiddleware': 543,
}

# Enable or disable downloader middlewares
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
DOWNLOADER_MIDDLEWARES = {
   'stock10jqka.middlewares.Stock10JqkaDownloaderMiddleware': 543,
}

# Enable or disable extensions
# See https://docs.scrapy.org/en/latest/topics/extensions.html
# EXTENSIONS = {
#    'scrapy.extensions.telnet.TelnetConsole': None,
# }

# Configure item pipelines
# See https://docs.scrapy.org/en/latest/topics/item-pipeline.html
ITEM_PIPELINES = {
    'stock10jqka.pipelines.Stock10JqkaPipeline': 300,
}

# Enable and configure the AutoThrottle extension (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/autothrottle.html
# AUTOTHROTTLE_ENABLED = True
# The initial download delay
# AUTOTHROTTLE_START_DELAY = 5
# The maximum download delay to be set in case of high latencies
# AUTOTHROTTLE_MAX_DELAY = 60
# The average number of requests Scrapy should be sending in parallel to
# each remote server
# AUTOTHROTTLE_TARGET_CONCURRENCY = 1.0
# Enable showing throttling stats for every response received:
AUTOTHROTTLE_DEBUG = False

# Enable and configure HTTP caching (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#httpcache-middleware-settings
# HTTPCACHE_ENABLED = True
# HTTPCACHE_EXPIRATION_SECS = 0
# HTTPCACHE_DIR = 'httpcache'
# HTTPCACHE_IGNORE_HTTP_CODES = []
# HTTPCACHE_STORAGE = 'scrapy.extensions.httpcache.FilesystemCacheStorage'
# REDIRECT_ENABLED = False  # 拒绝页面重定向，重定向的页面一般都是无法解析的，因为页面元素不一样
LOG_LEVEL = 'WARNING'
mysql_config = {  # 如果变量名为全大写，则可以在爬虫框架中的任何地方以settings.get(变量名的方式)获取settings.py中的变量值，小写则不行
    'host': 'localhost',
    'port': 3306,
    'user': 'root',
    'passwd': '111111',
    'charset': 'utf8mb4',
    'db': "stocks",
}
COLUMNS_STOCK = {  # 如果变量名为全大写，则可以在爬虫框架中的任何地方以settings.get(变量名的方式)获取settings.py中的变量值
    # 'code': 'varchar(9)',  # 9个字符长度，mysql8版本指的是字符长度（中英文都视作一个字符），早期版本可能指的是字节长度，
    # 表中不需要code列，因为表是以code命名的。
    'date': 'date',  # 主键
    '评分': 'float',
    '击败/%': 'tinyint',  # (-128,127)
    '短期趋势': 'varchar(50)',  # 0-255字节
    '中期趋势': 'varchar(50)',
    '长期趋势': 'varchar(50)',
    '技术面': 'float',
    '资金面': 'float',
    '消息面': 'float',
    '行业面': 'float',
    '基本面': 'float',
    '评级': 'tinyint',
    '地域': 'varchar(20)',
    '设计概念': 'varchar(500)',
    '主营业务': 'varchar(1000)',
    '每股净资产/元': 'float',
    '每股收益/元': 'float',
    '净利润/亿元': 'float',
    '净利润增长率/%': 'float',
    '营业收入/亿元': 'float',
    '每股现金流/元': 'float',
    '每股公积金/元': 'float',
    '每股未分配利润/元': 'float',
    '总股本/亿': 'float',
    '流通股/亿': 'float'
}
COLUMNS_STOCK_ITEM = {
    # mysql数据库中的列对应的item中的字段名
    'date': 'date',  # 主键
    '评分': '诊股_评分',
    '击败/%': '诊股_击败',
    '短期趋势': '诊股_短期趋势',
    '中期趋势': '诊股_中期趋势',
    '长期趋势': '诊股_长期趋势',
    '技术面': '诊股_技术面',
    '资金面': '诊股_资金面',
    '消息面': '诊股_消息面',
    '行业面': '诊股_行业面',
    '基本面': '诊股_基本面',
    '评级': '评级',
    '地域': '地域',
    '设计概念': '设计概念',
    '主营业务': '主营业务',
    '每股净资产/元': '每股净资产',
    '每股收益/元': '每股收益',
    '净利润/亿元': '净利润',
    '净利润增长率/%': '净利润增长率',
    '营业收入/亿元': '营业收入',
    '每股现金流/元': '每股现金流',
    '每股公积金/元': '每股公积金',
    '每股未分配利润/元': '每股未分配利润',
    '总股本/亿': '总股本',
    '流通股/亿': '流通股'
}
COLUMNS_NEW = {
    # mysql数据库中的列对应的item中的字段名相同，不用再单独配置
    # 'code': 'varchar(9)',
    'date': 'datetime',  # 主键
    '标题': 'varchar(100)',  # 主键s
    'url': 'varchar(200)',  # 可能包含特殊字符，不宜作为主键
    '内容': 'TEXT'  # 0-65535字节
}

NEO4J_CONFIG = {

}



"""
本模块使用iapws-97水及水蒸汽参数计算公式计算
"""
import iapws
from iapws.iapws97 import IAPWS97
from iapws.iapws95 import IAPWS95


def get_p_by_th(t, h):
    """
    根据h和t求p

    :param h: 焓，单位为 kJ/kg
    :param t: 温度，单位为 ℃
    :return: 压力，单位为 MPa
    """
    t_K = t + 273.15
    # 因为IAPWS97不能根据 焓 和 温度 求 压力，这里借助IAPWS95进行查找求解
    # 即多次计算 get_h_by_pt，对比计算的焓值与给定焓值相等则返回对应的p
    p = IAPWS95(h=h, T=t_K).P
    h_1 = get_h_by_pt(p=p, t=t)
    if h_1 > h:  # 温度不变时，压力越大，焓越低，考虑绝热管道的蒸汽流动，等焓，pt都降低
        # 说明估计的p偏小
        p_min = p
        p_max = p + 0.01
        while get_h_by_pt(p=p_max, t=t) > h:
            p_max = p_max + 0.01

    else:
        # 说明估计的p偏大
        p_max = p
        p_min = p - 0.01
        while get_h_by_pt(p=p_min, t=t) < h:
            p_min = p_min - 0.01
    # 真实的压力值位于p_min和p_max之间，二分法求解
    p_ave = (p_min + p_max) / 2
    h0 = get_h_by_pt(p=p_ave, t=t)
    while abs(h0 - h) > 0.000000001 and p_max - p_min > 0.000000001:
        if h0 > h:
            p_min = p_ave
        else:
            p_max = p_ave
        p_ave = (p_min + p_max) / 2
        h0 = get_h_by_pt(p=p_ave, t=t)

    return p_ave


def get_h_by_pt(p, t):
    """
    根据p和t求h

    :param p: 压力，单位为 MPa
    :param t: 温度，单位为 ℃
    :return: 焓，单位为 kJ/kg
    """
    h2o = IAPWS97(P=p, T=t + 273.15)
    return h2o.h


def get_s_by_pt(p, t):
    """
    根据p和t求熵

    :param p: 压力，单位为 MPa
    :param t: 温度，单位为 ℃
    :return: 焓，单位为 kJ/kg
    """
    h2o = IAPWS97(P=p, T=t + 273.15)
    return h2o.s


def get_t_by_hp(h, p):
    """
    根据h和p求t

    :param h: 焓，单位为 kJ/kg
    :param p: 压力，单位为 MPa
    :return: 温度，单位为 K
    """
    h2o = iapws.iapws97.IAPWS97_Ph(P=p, h=h)
    return h2o.T - 273.15


def get_h_by_px(p, x):
    """
    根据p和x求h

    :param p: 压力，单位为 MPa
    :param x: 蒸汽干度
    :return: 焓，单位为 kJ/kg
    """
    h2o = iapws.iapws97.IAPWS97_Px(P=p, x=x)
    return h2o.h


def get_t_by_p(p):
    """
    根据压力求对应的饱和温度

    :param p: MPa
    :return:
    """
    return iapws.iapws97._TSat_P(p) - 273.15


def get_p_by_t(t):
    """
        根据温度求对应的饱和压力

        :param t: ℃
        :return:
        """
    return iapws.iapws97._PSat_T(t + 273.15)


H_PT = get_h_by_pt
S_PT = get_s_by_pt
P_T = get_p_by_t
T_P = get_t_by_p
P_TH = get_p_by_th

if __name__ == "__main__":
    print(f"{H_PT(10, 500)=}")
    print(f"{S_PT(10, 500)=}")
    print(f"{T_P(1)=}")
    print(f"{P_T(200)=}")
    print(f"{P_TH(500, 3300)=}")



# -*- coding: utf-8 -*-

# Define here the models for your spider middleware
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/spider-middleware.html

from scrapy import signals
from scrapy.exceptions import IgnoreRequest
from scrapy.http import HtmlResponse
from selenium import webdriver


class Stock10JqkaSpiderMiddleware(object):
    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the spider middleware does not modify the
    # passed objects.

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_spider_input(self, response, spider):
        # Called for each response that goes through the spider
        # middleware and into the spider.

        # Should return None or raise an exception.
        return None

    def process_spider_output(self, response, result, spider):
        # Called with the results returned from the Spider, after
        # it has processed the response.

        # Must return an iterable of Request, dict or Item objects.
        for i in result:
            yield i

    def process_spider_exception(self, response, exception, spider):
        # Called when a spider or process_spider_input() method
        # (from other spider middleware) raises an exception.

        # Should return either None or an iterable of Request, dict
        # or Item objects.
        pass

    def process_start_requests(self, start_requests, spider):
        # Called with the start requests of the spider, and works
        # similarly to the process_spider_output() method, except
        # that it doesn’t have a response associated.

        # Must return only requests (not items).
        for r in start_requests:
            yield r

    def spider_opened(self, spider):
        spider.logger.info('Spider opened: %s' % spider.name)


class Stock10JqkaDownloaderMiddleware(object):
    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the downloader middleware does not modify the
    # passed objects.
    # 将driver创建在中间件的初始化方法中，适用于项目中只有一个爬虫，设置好后需要在settings.py中启用
    # def __init__(self):
    # 在scrapy中创建driver对象，尽可能少的创建该对象
    # self.browser = webdriver.Edge()
    url_already_requested = []

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request, spider):
        # Called for each request that goes through the downloader
        # middleware.

        # Must either:
        # - return None: continue processing this request
        # - or return a Response object
        # - or return a Request object
        # - or raise IgnoreRequest: process_exception() methods of
        #   installed downloader middleware will be called
        if request.url in self.url_already_requested:
            raise IgnoreRequest("重复请求")
        self.url_already_requested.append(request.url)
        spider.browser.get(request.url)
        origin_code = spider.browser.page_source
        print(f"爬取网页：[{request.url}]......")
        res = HtmlResponse(url=request.url, encoding='utf8', body=origin_code, request=request)
        return res
        # return None

    def process_response(self, request, response, spider):
        # Called with the response returned from the downloader.

        # Must either;
        # - return a Response object
        # - return a Request object
        # - or raise IgnoreRequest
        return response

    def process_exception(self, request, exception, spider):
        # Called when a download handler or a process_request()
        # (from other downloader middleware) raises an exception.

        # Must either:
        # - return None: continue processing this exception
        # - return a Response object: stops process_exception() chain
        # - return a Request object: stops process_exception() chain
        print(f"-----------------重复请求：[{request.url}]")

    def spider_opened(self, spider):
        spider.logger.info('Spider opened: %s' % spider.name)



import copy
import random
from collections import OrderedDict
from typing import Optional

import numpy as np
import optuna
import pandas as pd
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from torch.utils.tensorboard import SummaryWriter

from yangke.base import get_settings, YkDict
from yangke.common.config import logger
from yangke.common.fileOperate import read_csv_ex
import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


class DataSetFitting(Dataset):
    def __init__(self, data_source, x_title=None, y_title=None, mean=None, std=None):
        """
        构建数据集

        :param data_source: 数据源，可以为settings.yaml文件配置，也可以是pd.DataFrame对象
        :param x_title: 输入参数的列标题，如果从配置文件加载，则可以为None
        :param y_title: 输出参数的列标题，如果从配置文件加载，则可以为None
        """
        self.dataframe = None
        if isinstance(data_source, pd.DataFrame):
            self.dataframe = data_source
            self.x_title = x_title or []
            self.y_title = y_title or []
        elif isinstance(data_source, YkDict):
            files = data_source.get_settings("dataset.data_file")
            self.normal_method = data_source.get_settings("dataset.normalization") or "z-score"
            input_set = data_source.get("input")
            self.x_title = [item["name"] for item in input_set]
            self.x_range = [None if item.get("range") is None else eval(item["range"]) for item in input_set]
            del input_set
            output_set = data_source.get_settings("output.para")
            if len(output_set) > 0 and isinstance(output_set[0], str):
                self.y_title = output_set
                self.y_range = None
            else:
                self.y_title = [item["name"] for item in output_set]
                self.y_range = [None if item.get("range") is None else eval(item["range"]) for item in output_set]
            for file in files:
                ext = file.get("type")
                filename = file.get("name")
                if not os.path.exists(filename):
                    continue
                if ext == "csv":
                    data_frame = read_csv_ex(filename)
                else:
                    data_frame = pd.read_excel(filename)
                if self.dataframe is not None and data_frame is not None:
                    self.dataframe = pd.concat([self.dataframe, data_frame], axis=0, ignore_index=True)
                else:
                    self.dataframe = self.dataframe or data_frame

            # 删除某些行
            self.drop_by_condition(data_source.get_settings("dataset.drop"))

        # 仅保留x,y相关的数据列，删除其他列
        titles = self.x_title.copy()
        titles.extend(self.y_title)
        if self.dataframe.shape[0] < 1:
            logger.warning("传入的数据文件中数据为空或数据文件不存在")
        self.dataframe = self.dataframe[titles]

        # 数据标准化
        self.mean = mean
        self.std_range = std
        self.dataframe_std = None
        self.set_standard()

    def set_standard(self, mean=None, std_range=None, normal_method=None):
        """
        设置数据的归一化参数，在初始化部分数据时，直接从部分数据集得到的归一化参数可能出现偏差，这里提供一种外部设置的方法。
        不建议中途更改数据归一化方法，否则可能导致模型预测结果很离谱。

        :param mean: 参数的平均值
        :param std_range: 参数的标准差或振幅
        :param normal_method: 数据归一化方法，支持"z-score"和"min-max"
        :return:
        """
        if normal_method is None or normal_method == self.normal_method:
            if mean is None and std_range is None:
                if self.normal_method == "z-score":
                    self.mean = self.mean or self.dataframe.mean()
                    self.std_range = self.std_range or self.dataframe.std()
                else:
                    if self.y_range is not None:
                        x_min = [x[0] for x in self.x_range]
                        y_min = [y[0] for y in self.y_range]
                        x_max = [x[1] for x in self.x_range]
                        y_max = [y[1] for y in self.y_range]
                        x_min.extend(y_min)
                        all_min = np.asarray(x_min)
                        x_max.extend(y_max)
                        all_max = np.asarray(x_max)
                        title = copy.deepcopy(self.x_title)
                        title.extend(self.y_title)
                        self.mean = pd.Series(data=(all_min + all_max) / 2, index=title)
                        self.std_range = pd.Series(data=(all_max - all_min) / 2, index=title)
            else:
                self.mean = mean
                self.std_range = std_range
            if self.mean is not None and self.std_range is not None:
                self.dataframe_std = self.standard(self.dataframe)
            else:
                self.dataframe_std = None
        else:
            self.normal_method = normal_method
            self.set_standard()

    def standard(self, df):
        """
        数据标准化

        :param df: Dataframe(num, input_and_output_dim)
        :return:
        """
        # Dataframe和Series运算的前提是：Dataframe的列标题和Series的索引名必须一一对应，而不仅仅看维度，否则可能会出错，
        return (df - self.mean) / self.std_range  # mean Series (input_and_output_dim,)

    def standard_reverse(self, out):
        """
        将预测结果反标准化

        :param out:
        :return:
        """
        return out * self.std_range[self.y_title] + self.mean[self.y_title]

    def __getitem__(self, index):
        """
        DataSet子类必须实现的方法，用于根据索引返回一条数据，数据类型需要是Tensor

        :param index:
        :return:
        """
        single_item = self.dataframe_std.iloc[index, :]
        x = torch.from_numpy(single_item[self.x_title].to_numpy()).to(torch.float32)
        y = torch.from_numpy(single_item[self.y_title].to_numpy()).to(torch.float32)
        return x, y

    def __len__(self):
        """
        DataSet子类必须实现的方法，用于获取DataSet的大小

        :return:
        """
        return self.dataframe.shape[0]

    def get_size(self):
        return self.dataframe.shape[0]

    def drop_by_condition(self, conditions):
        for cond in conditions:
            if list(cond.keys())[0] == "or":
                for co in cond.get("or"):
                    if "<=" in co:
                        title, value = tuple(co.split("<="))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] > value]  # 删除小于的行 = 保留大于的行
                    elif ">=" in co:
                        title, value = tuple(co.split(">="))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] < value]  # 删除小于的行 = 保留大于的行
                    elif "<" in co:
                        title, value = tuple(co.split("<"))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] >= value]  # 删除小于的行 = 保留大于的行
                    elif ">" in co:
                        title, value = tuple(co.split(">"))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] <= value]  # 删除大于的行 = 保留小于的行

    def split_set(self, part1, part2=None, part3=None):
        """
        按照指定的比例将数据集分割，一般用于将总体数据集分割为训练集，测试集，验证集等

        :param part1:
        :param part2:
        :param part3:
        :return:
        """
        if part3 is not None:
            assert part1 + part2 + part3 == 1, "数据集比例之和不为1"
            size = self.get_size()
            size1, size2 = int(part1 * size), int(part2 * size)
            set1, set2, set3 = torch.utils.data.random_split(self, [size1, size2, size - size1 - size2])
            return set1, set2, set3
        elif part2 is not None:
            if part1 + part2 < 1:
                return self.split_set(part1, part2, 1 - part1 - part2)
            else:
                size = self.get_size()
                size1 = int(part1 * size)
                return torch.utils.data.random_split(self, [size1, size - size1])
        else:
            size = self.get_size()
            size1 = int(part1 * size)
            return torch.utils.data.random_split(self, [size1, size - size1])


class DataFitterNet(torch.nn.Module):
    def __init__(self, settings1, trial=None):
        """
        一个用于数据拟合的神经网络类库，神经网络架构通过settings.yaml文件进行配置

        :param settings1:
        :param trial: 使用optuna超参数调优时会传入该参数
        """
        super(DataFitterNet, self).__init__()
        cfg = settings1.get("networks")  # 获取神经网络结构信息
        self.settings = settings1
        self.trial = trial
        self.cfg = cfg
        self.in_num = 0  # 输入层神经元个数，对应输入参数的个数
        self.out_num = 1
        train_settings = settings1.get_settings("networks.train") or {}
        self.lr = self.get_para_of_optuna_str(train_settings.get("learning_rate") or 1e-3, "lr")
        self.epochs = int(train_settings.get("epochs") or 10)
        self.batch_size = self.get_para_of_optuna_str(train_settings.get("batch_size") or 64, "batch_size")
        if settings1.get_settings("networks.loss_fn.type") == "MSE":
            self.loss_fn = nn.MSELoss(reduction="mean")
        elif settings1.get_settings("networks.loss_fn.type").lower() == "max":
            from yangke.stock.prediction.pytorch_pred_stock import MAXLoss
            self.loss_fn = MAXLoss()
        else:
            self.loss_fn = nn.CrossEntropyLoss()
        self.mean: Optional[pd.Series] = None
        self.std: Optional[pd.Series] = None
        self.normal_method = None  # 数据归一化方法
        self.x_title = []
        self.y_title = []
        self.max_err = 0
        self.average_err = 0
        _cell_num_last_layer = 0  # 循环中记录上一层神经网络的输出个数
        i = 1
        layer_dict = OrderedDict()
        for layer in cfg.get("layers"):  # 从配置中拿到神经网络各层的信息，进而构建神经网络
            _type = self.get_para_of_optuna_str(layer.get("type"), f"layer_type{i}")
            _cell_num = 1
            if _type == "input":  # 输入/输出层的神经元个数可以根据输入/输出参数个数自动确定
                _cell_num = layer.get("cell_num") or 10
                if layer.get("cell_num") == "auto" or layer.get("cell_num") is None:
                    self.in_num = len(settings1.get("input"))
                    _cell_num = self.in_num
                else:
                    self.in_num = int(layer.get("cell_num") or 10)
                    _cell_num = self.in_num
            elif _type == "linear":
                bias = layer.get("bias") or True
                _cell_num = self.get_para_of_optuna_str(layer.get("cell_num") or 10, f"cell_num{i}")
                layer_dict[f"layer{i}"] = nn.Linear(_cell_num_last_layer, _cell_num, bias=bias)
            elif _type == "relu":
                layer_dict[f"layer{i}"] = nn.ReLU()
                _cell_num = _cell_num_last_layer  # 激活函数输入输出个数不变
            # elif _type == 'softmax':
            #     self.layers.append(nn.Softmax())
            elif _type == "sigmoid":
                layer_dict[f"layer{i}"] = nn.Sigmoid()
                _cell_num = _cell_num_last_layer
            elif _type == "tanh":
                layer_dict[f"layer{i}"] = nn.Tanh()
                _cell_num = _cell_num_last_layer
            elif _type == "dropout":
                p = self.get_para_of_optuna_str(layer.get('rate'), "dropout_rate")
                layer_dict[f"layer{i}"] = nn.Dropout(p=p)
                _cell_num = _cell_num_last_layer
            elif _type.lower() == "rbf":
                _cell_num = layer.get("cell_num") or 10
                self.in_num = len(settings1.get("input"))
                self.out_num = len(settings1.get("output"))
                logger.error("暂不支持径向基神经网络")
            elif _type.lower() == "none":  # 空层
                continue
            elif _type == "output":  # 输入/输出层的神经元个数可以根据输入/输出参数个数自动确定
                if settings1.get_settings("output.type") == "single_output":
                    if layer.get("cell_num") != "auto" and int(layer.get("cell_num")) != 1:
                        logger.warning("数据集使用single_output，但神经网络模型输出层单元个数不为1，忽略神经网络输出单元个数设置！")
                    self.out_num = 1
                elif layer.get("cell_num") == "auto" or layer.get("cell_num") is None:
                    self.out_num = len(settings1.get("output"))
                else:
                    self.out_num = layer.get("cell_num")
                bias = layer.get("bias") or True
                layer_dict[f"layer{i}"] = nn.Linear(_cell_num_last_layer, self.out_num, bias=bias)
            else:
                logger.warning(f"发现不支持的layer类型：{_type}")

            _cell_num_last_layer = _cell_num
            i = i + 1

        self.net = nn.Sequential(layer_dict)
        weight_decay = self.get_para_of_optuna_str(settings1.get_settings("networks.optimizer.weight_decay") or 0,
                                                   "weight_decay")
        if settings1.get_settings("networks.optimizer.type") == "adam":
            self.optimizer = torch.optim.Adam(self.parameters(), lr=self.lr, weight_decay=weight_decay)
        else:
            self.optimizer = torch.optim.SGD(self.parameters(), lr=self.lr, weight_decay=weight_decay)

    def forward(self, x):
        out = self.net(x)
        return out

    def set_standard(self, mean, std, x_title, y_title, normal_method="z-score"):
        self.mean = mean
        self.std = std
        self.x_title = x_title
        self.y_title = y_title
        self.normal_method = normal_method

    def standard_reverse(self, tensor, flag="y"):
        """
        将预测结果反标准化

        :param tensor: 反标准化的数据
        :param flag: 反标准化的变量类型，x表示对输入变量进行反标准换，y表示对输出变量
        :return:
        """
        if len(self.y_title) != 1:
            logger.error("暂不支持输出变量个数大于1的神经网络")
        if flag == "x":
            para = tensor.cpu().data.numpy()  # 必须加.data，否则在需要求导的Tensor上直接调用numpy()方法会出错
            std = self.std[self.x_title].to_numpy()
            mean = self.mean[self.x_title].to_numpy()
        elif flag == "y":
            para = tensor.cpu().data.numpy()
            std = self.std[self.y_title].to_numpy()
            mean = self.mean[self.y_title].to_numpy()
        else:  # "all"包括x和y
            para = tensor.cpu().data.numpy()
            std = self.std.to_numpy()
            mean = self.mean.to_numpy()
        result = para * std + mean
        result = torch.from_numpy(result)
        return result

    def prediction(self, x_input):
        """
        预测指定输入参数

        :param x_input: x.shape = torch.Size([batch_size, input_dim])，也可以是长度为input_dim的列表或ndarray
        :return:
        """
        mean = torch.from_numpy(self.mean[self.x_title].to_numpy())
        std = torch.from_numpy(self.std[self.x_title].to_numpy())
        if isinstance(x_input, list):
            x_input = torch.from_numpy(np.asarray(x_input))
        if isinstance(x_input, np.ndarray):
            x_input = torch.from_numpy(x_input)
        x_input = (x_input - mean) / std  # 输入数据归一化
        x_input = x_input.to(torch.float32)
        self.eval()
        y = self.net(x_input)
        return self.standard_reverse(y)

    def save_yk(self, path):
        checkpoint = {
            'normal_method': self.normal_method,
            'mean': self.mean,
            'std': self.std,
            'x_title': self.x_title,
            'y_title': self.y_title,
            'model_state_dict': self.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'max_err': self.max_err,
            'average_err': self.average_err,
            'settings': self.settings
        }
        torch.save(checkpoint, path)

    @staticmethod
    def load_yk(path):
        if path is None or not os.path.exists(path):
            logger.debug("神经网络模型文件不存在！")
            return None
        checkpoint = torch.load(path)
        model1 = DataFitterNet(checkpoint.get("settings"))
        model1.set_standard(checkpoint['mean'], checkpoint['std'], checkpoint['x_title'], checkpoint['y_title'],
                            checkpoint['normal_method'])
        model1.max_err = checkpoint['max_err']
        model1.average_err = checkpoint['average_err']
        model1.load_state_dict(checkpoint['model_state_dict'])
        model1.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        return model1

    def __str__(self):
        str1 = super(DataFitterNet, self).__str__()
        return str1[:-1] + "  (Optimizer): " + self.optimizer.__str__().replace("\n", "\n  ", 1).replace("\n",
                                                                                                         "\n  ") + "\n)"

    def device(self):
        return next(self.parameters()).device

    def train_loop(self, data_loader):
        size = len(data_loader.dataset)
        self.train()
        loss = 0
        is_tqdm = self.settings.get_settings("print.tqdm")
        if is_tqdm:
            from tqdm import tqdm
            pbar = tqdm(data_loader)
        else:
            pbar = data_loader
            pbar.set_set_description = print
        for batch, (x, y) in enumerate(pbar):  # 这里的(x, y)是归一化后的数据
            x = x.to(self.device())
            y = y.to(self.device())
            pred = self.net(x)
            loss = self.loss_fn(pred, y)
            self.optimizer.zero_grad()  # 每一次训练梯度要清零
            loss.backward()
            self.optimizer.step()

            if batch % 10 == 0:
                loss, current = loss.item(), batch * len(x)
                pbar.set_description(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}-{self.y_title}]")

        return loss

    def test_loop(self, data_loader):
        num_batches = len(data_loader)
        size = len(data_loader.dataset)
        test_loss = 0
        self.eval()
        relative_err = 0
        max_err = 0
        with torch.no_grad():
            for x, y in data_loader:
                x = x.to(self.device())
                y = y.to(self.device())
                pred = self.net(x)
                test_loss += self.loss_fn(pred, y).item()
                pred = self.standard_reverse(pred)
                y = self.standard_reverse(y)
                # max_err = max(max_err, (abs(pred-y)/abs(y)).max(0))  # tensor.max()返回最大值,tensor.max(0)返回最大值和索引
                max_err = max(max_err, (abs(pred - y) / abs(y)).max())
                relative_err = ((pred - y).abs() / y.abs()).sum() + relative_err

        test_loss /= num_batches
        relative_err = relative_err / size
        logger.debug(f"测试集预测结果({self.y_title[0]}): 平均损失为{test_loss:>8f}")
        return relative_err, max_err

    def start_train(self, train_set, test_set, writer=None):
        """
        开始模型训练
        :param train_set:
        :param test_set:
        :param writer: tensorboard的记录器
        :return:
        """
        relative_err = []  # 预测结果的相对误差
        if writer is None:
            writer = SummaryWriter(f'./runs/{self.y_title[0]}')  # 可视化数据存放在这个文件夹
        writer.add_graph(self, torch.rand([1, self.in_num], device=self.device()))
        train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, shuffle=True)
        test_loader = DataLoader(dataset=test_set, batch_size=self.batch_size, shuffle=True)
        for t in range(self.epochs):
            logger.debug(f"-------------------------Epoch {t + 1}-------------------------------")
            train_loss = self.train_loop(train_loader)
            current_err, max_err = self.test_loop(test_loader)
            relative_err.append(current_err)
            logger.debug(f"最大相对误差为{max_err.item():>8f}，平均相对误差为：{current_err:>8f}")
            # ------------------------- 损失曲线图 ---------------------------------
            writer.add_scalar('train/损失', train_loss, t)
            writer.add_scalar('test/平均误差', current_err, t)
            writer.add_scalar('test/最大误差', max_err, t)
            # ------------------------- 损失曲线图 ---------------------------------
            # ------------------------- 权重直方图 ---------------------------------
            for i, (name, param) in enumerate(self.named_parameters()):
                if 'bn' not in name:
                    writer.add_histogram(name + "_data", param, t)
                    writer.add_histogram(name + "_grad", param.grad, t)
            # ------------------------- 权重直方图 ---------------------------------
            try:
                if current_err < 0.01 and max_err < 0.07 and t >= 20:
                    self.max_err = max_err
                    self.average_err = current_err
                    break
            except IndexError:  # 当relative_err列表长度小于3时，会报IndexError错误，此处忽略错误继续训练
                pass
            self.max_err = max_err
            self.average_err = current_err
            if self.trial is not None:
                self.trial.report(self.average_err, t)
                if self.trial.should_prune():
                    raise optuna.TrialPruned
        logger.debug(f"测试集预测结果的相对误差随epoch的变化为：{[x.item() for x in relative_err]}")
        writer.close()
        return self.average_err, self.max_err

    def get_para_type(self, p):
        if isinstance(p, list):
            result = "int"
            for p_ in p:
                if self.get_para_type(p_) == "float":
                    result = "float"
                elif self.get_para_type(p_) == "str":
                    result = "str"
            return result
        else:
            if str(p).strip().isnumeric():  # 小数的isnumeric()方法返回的是false
                return "int"
            else:
                try:
                    _ = eval(p)
                    return "float"
                except NameError:
                    return "str"

    def get_para_of_optuna_str(self, string, name):
        """
        处理设置项中的Optuna优化项，当设置项的值以Optuna开头时，则自动返回Optuna的采样值。

        :param string: 设置项的值
        :param name: 变量名，不影响返回值，但是Optuna工具根据该值识别是哪个变量的采样值，从而给出采样值
        :return:
        """
        result = 0
        if str(string).startswith("optuna"):
            p1 = string.replace("optuna(", "").replace(")", "")
            if "[" not in p1:
                p_list = p1.split(",")
                type1 = self.get_para_type(p_list)
                if type1 == "float":
                    step = None if len(p_list) == 2 else float(p_list[2])
                    result = self.trial.suggest_float(name, float(p_list[0]), float(p_list[1]), step=step)
                elif type1 == "int":
                    step = None if len(p_list) == 2 else int(p_list[2])
                    result = self.trial.suggest_int(name, int(p_list[0]), int(p_list[1]), step=step)
            else:
                p_list = eval(p1)
                result = self.trial.suggest_categorical(name, p_list)
        else:
            result = string
        return result


class OptunaModel:
    def __init__(self, settings=None):
        if settings is None:
            settings = YkDict({})
        self.settings = copy.deepcopy(settings)
        self.settings["networks"] = copy.deepcopy(settings.get_settings("optuna.networks"))
        self.n_trials = int(settings.get_settings("optuna.n_trials") or 10)
        self.device = 'cpu'
        self.mean = None
        self.std = None
        self.x_title = None
        self.y_title = None
        self.train_set = None
        self.test_set = None

    def optimize(self):
        study_name = "model_study20220525"
        study = optuna.create_study(study_name=study_name, direction="minimize",
                                    storage=f'sqlite:///{study_name}.db', load_if_exists=True)
        study.optimize(self.objective, n_trials=self.n_trials)
        trial = study.best_trial
        logger.debug(f"最优模型的损失为{trial.value}")
        logger.debug(f"最优模型的参数为{trial.params}")
        df = study.trials_dataframe()
        print(df)
        self.visualization(study_name)

    @staticmethod
    def visualization(study_name):
        study = optuna.create_study(study_name=study_name, direction="minimize",
                                    storage=f'sqlite:///{study_name}.db', load_if_exists=True)
        optuna.visualization.plot_contour(study).show()  # 参数的等高线图
        optuna.visualization.plot_optimization_history(study).show()  # 优化历史曲线
        optuna.visualization.plot_param_importances(study).show()  # 参数的重要性
        optuna.visualization.plot_intermediate_values(study).show()  # Trails的学习曲线
        optuna.visualization.plot_parallel_coordinate(study).show()  # 高纬度参数的关系，看不懂
        optuna.visualization.plot_slice(study).show()  # 可视化独立参数

    def start_train(self, train_set, test_set):
        self.train_set = train_set
        self.test_set = test_set
        return self.optimize()

    def objective(self, trial):
        model = DataFitterNet(self.settings, trial).to(self.device)
        model.set_standard(self.mean, self.std, self.x_title, self.y_title)
        logger.debug("try model:")
        print(model)
        loss, max_err = model.start_train(self.train_set, self.test_set)
        max_err = max_err.item() if isinstance(max_err, torch.Tensor) else max_err
        loss = loss.item() if isinstance(loss, torch.Tensor) else loss
        trial.set_user_attr('max_err', max_err)
        return loss

    def to(self, device):
        self.device = device
        return self

    def set_standard(self, mean, std, x_title, y_title, *args):
        self.mean = mean
        self.std = std
        self.x_title = x_title
        self.y_title = y_title


def train_model(settings1=None):
    """
    按照settings.yaml文件训练神经网络模型，并保存到settings.yaml指定的位置

    如果settings.networks没有定义具体的网络模型，但是settings.optuna.networks中定义了网络模型的寻优空间参数，则会自动使用
    Optuna库对模型超参数进行寻优，并将寻优结果得到的最有网络模型定义写入settings.yml文件中。再次调用该方法，则会使用寻优得到的
    网络模型架构进行数据拟合，将模型权值信息写入settings.output.save_path指定的路径中。

    如果settings.networks定义了具体的网络模型，则该方法只对定义的模型进行训练，将训练后的权值信息写入settings.output.save_path
    指定的路径中。

    如果该方法同时训练多个单输出模型，则模型架构的优化只会进行一次，即第一个模型架构的优化结果会应用于后续的多个单输出数据拟合任务。

    :return:
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    logger.info(f"使用{device}进行计算")
    if settings1 is None:
        settings1 = get_settings()

    if settings1.get("networks") is not None:  # 如果存在networks的定义就直接使用该定义
        optimize_model = False
    elif settings1.get("optuna") is not None:  # 如果存在optuna的设置，就使用该设置获取最有网络模型定义
        optimize_model = True
    else:
        logger.error("未找到模型设置信息，请确认！")
        return
    _settings_ = copy.deepcopy(settings1)
    output = _settings_.get("output") or {}
    if output.get("type") == "single_output":
        paras = copy.deepcopy(output.get("para") or {})
        for i, para in enumerate(paras):
            output["para"] = [para]
            output["save_path"] = settings1.get_settings("output.save_path")[i]
            model = OptunaModel(_settings_).to(device) if optimize_model else DataFitterNet(_settings_).to(device)
            logger.debug(model)
            dataset_all = DataSetFitting(_settings_)  # 获取所有的数据集
            model.set_standard(dataset_all.mean, dataset_all.std_range, dataset_all.x_title, dataset_all.y_title,
                               dataset_all.normal_method)  # 将数据集的数据归一化信息添加到模型中，方便以后使用
            part1 = float(_settings_.get_settings("dataset.data_training.proportion"))
            part2 = float(_settings_.get_settings("dataset.data_test.proportion"))
            train_set, test_set = dataset_all.split_set(part1, part2)  # 均已是归一化后的数据
            model.start_train(train_set, test_set)
            if not optimize_model:
                model.save_yk(output["save_path"])
            else:
                model.update_settings_file("settings.yml")
                train_model()


def re_train(data_file, settings1=None):
    """
    在现有模型基础上再次训练，该方法对数据的归一化方法必须与初次训练时一致，因此本方法会忽略settings1中的数据归一化相关设置，而自动采用原模型中
    的数据归一化方法

    :param data_file: 数据文件,csv文件
    :param settings1: 可以直接传入settings字典，则该方法不会再从外部yaml文件读取设置信息
    :return:
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    logger.info(f"使用{device}再次训练")
    if settings1 is None:
        settings1 = get_settings()
    _settings_ = copy.deepcopy(settings1)
    if isinstance(data_file, list):
        _data_file = []
        for file in data_file:
            _data_file.append({"name": file, "type": "csv"})
        _settings_["dataset"]["data_file"] = _data_file
    else:
        _settings_["dataset"]["data_file"] = [{"name": data_file, "type": "csv"}]
    output = _settings_.get("output") or {}
    if output.get("type") == "single_output":
        paras = copy.deepcopy(output.get("para") or {})
        for i, para in enumerate(paras):
            output["para"] = [para]
            output["save_path"] = settings1.get_settings("output.save_path")[i]
            model = DataFitterNet.load_yk(output["save_path"])
            model.epochs = _settings_.get_settings("networks.train.epochs")  # 再次训练时的epochs和初次训练时的一般不相同
            logger.debug(model)
            dataset_all = DataSetFitting(_settings_)  # 获取所有的数据集
            if _settings_.get_settings("dataset.normalization") != model.normal_method:
                logger.warning(f"再训练时数据归一化方法({_settings_.get_settings('dataset.normalization')})与"
                               f"模型初次训练时({model.normal_method})不一致，这可能导致预测结果错误")
                logger.warning(f"已强制切换为模型初次训练时使用的数据归一化方法！({model.normal_method})")
            dataset_all.set_standard(model.mean, model.std, model.normal_method)
            part1 = 0.95
            part2 = 0.05
            train_set, test_set = dataset_all.split_set(part1, part2)  # 均已是归一化后的数据
            model.start_train(train_set, test_set)

            model.save_yk(output["save_path"])


def filter_err_condition(model, data_file=None):
    """
    将model在整个数据集上进行测试，过滤小偏差的点，删选出偏差较大的点，一般后续对偏差较大的点继续学习

    :param model:
    :param data_file:
    :return:
    """
    settings = get_settings()
    if data_file is not None:
        settings["dataset"]["data_file"] = [{"name": data_file, "type": "csv"}]
    _settings_ = copy.deepcopy(settings)
    dataset_all = DataSetFitting(_settings_)
    dataset_all.set_standard(model.mean, model.std_range)


def setup_seed(seed=10):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def test():
    model = DataFitterNet.load_yk(r"D:\lengduan\data\model_p2.dat")
    args = {
        "power": 400,
        "flow_heat": 0,
        "p_env": 98,
        "t_env": 30,
        "humid": 0.60,
        "p_gas": 3.8,
        "t_gas": 18,
        "flow_fgh": 40,
        "flow_tca": 100,
        "flow_oh": 3,
        "flow_rh": 0
    }
    x = [v for k, v in args.items()]
    pump, fun = 2, 3
    x.extend([pump, fun])
    p = model.prediction(x)
    print(p)


def deal_settings_path(settings, **kwargs):
    """
    当settings.yaml文件中使用了带变量的路径时，如"E:/热工院//retrain{unit}_0.csv",
    本方法将变量{unit}替换为指定的变量。

    :param settings:
    :return:
    """
    files = []

    for data_file in settings.get_settings("dataset.data_file"):
        pth = data_file["name"].replace(r"{unit}", "1")
        type1 = data_file["type"]
        files.append({"name": pth, "type": type1})
    settings["dataset"]["data_file"] = files
    return settings


if __name__ == "__main__":
    setup_seed()
    OptunaModel(None).visualization("model_study20220525")
    # settings1 = get_settings()
    # deal_settings_path(settings1)
    # train_model(settings1)
    # # settings["output"]["save_path"] = [r"D:\lengduan\data\model_p1.dat"]
    # # settings["output"]["para"] = ["背压"]
    # re_train(data_file=[
    #     r"D:\lengduan\data\retrain1_0.csv",
    #     r"D:\lengduan\data\retrain1_1.csv",
    #     r"D:\lengduan\data\retrain1_7.csv",
    #     r"D:\lengduan\data\retrain1_9.csv",
    #     r"D:\lengduan\data\retrain1_10.csv",
    #     r"D:\lengduan\data\retrain1_11.csv",
    # ], settings1=settings_1)
    # re_train(data_file=[
    #     r"D:\lengduan\data\retrain2_0.csv",
    #     r"D:\lengduan\data\retrain2_1.csv",
    #     r"D:\lengduan\data\retrain2_7.csv",
    #     r"D:\lengduan\data\retrain2_8.csv",
    #     r"D:\lengduan\data\retrain2_9.csv",
    #     r"D:\lengduan\data\retrain2_10.csv",
    #     r"D:\lengduan\data\retrain2_11.csv",
    # ], settings1=settings_1)



from scrapy import cmdline
import os, sys

# 创建scrapy项目的命令： python -m scrapy startproject <project-name>
# 进入爬虫项目目录：     cd <project-name>
# 初始化一个爬虫：       python -m scrapy genspider <爬虫名> <网站域名>
# 初始化另一个爬虫：     python -m scrapy genspider -t crawl jqka_spider "10jqka.com.cn"
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
cmdline.execute("scrapy crawl jqka_spider".split())  # 执行cmdline.execute()方法

# scrapy爬虫执行顺序
# 1. cmdline中的execute方法进入爬虫，初始化settings中的各类参数
# 2. crawler.py中启动CrawlerProcess，CrawlerProcess初始化SpiderLoader
# 2.1 SpiderLoader通过from_settings()加载settings中定义的spider，通过self._load_all_spiders()方法
# 2.2 加载用户在settings.py文件的SPIDER_MODULES中定义的文件夹中的所有spider
# 2.3 初始化pipelines.py中的自定义Pipeline
# 3. 开始爬取



from flask import Flask, request, jsonify, make_response, send_from_directory
from yangke.base import name_transfer
from flask_cors import *
from yangke.common.config import logger
import traceback
import os

download_func = None
example = None


def start_server_app(deal=None, app_name='yangke.web.flaskserver', instance_path=__file__,
                     instance_relative_config=True, host="0.0.0.0", port="5000", allow_action=None, use_action=True,
                     example_url=None):
    """
    启动Flask Server服务，服务启动后将调用用户传入的deal函数来对url请求进行处理，deal函数通过args参数拿到请求的参数字典。
    如果用户需要自定义deal函数处理http://127.0.0.1:5000这类没有参数的url，则应设置useAction为False，并在deal函数中对空字典进行判断处理。
    如果用户需要过滤允许的Action类型，则需要设置允许的操作类型列表allow_action。

    当use_action为True时，如果请求url不带Action参数，则该方法会自动返回提示信息。因此，传递给用户函数deal的参数字典中确保含有Action参数。
    当use_action为False时，即使请求url不带任何参数，也会调用deal函数，且将对应的参数字典或空字典传递给deal函数，deal函数需要判空。

    使用示例：
    #==============================================================================================
    def deal(args):
        action = args.get('Action')  # 因为下方use_action=True，所以这里的action必然有值，避免eval函数出错
        result = eval("{}(args)".format(action))
        return result

    start_server_app(deal, use_action=True)  # 启动服务器.
    #==============================================================================================

    :param use_action: 是否使用Action参数指定操作类型，不使用则完全由用户定义所有url响应
    :param deal: 用户指定的url处理方法
    :param app_name: Flask app名，一般不需要也不能更改，Flask根据该参数查找模板和静态文件等东西
    :param instance_path:
    :param instance_relative_config:
    :param host:
    :param port:
    :param allow_action:  允许用户调用的方法
    :param example_url: 示例http访问请求，当用户访问出错时提示用户
    :return:
    """
    app = Flask(app_name, instance_path=instance_path, instance_relative_config=instance_relative_config)
    app.config['JSON_AS_ASCII'] = False  # 让Flask返回的字符串可以正确显示中文，默认显示的是utf-8编码
    CORS(app)
    global example
    example = example_url

    @app.route('/', methods=['GET', 'POST'])
    def start_service():
        action = request.args.get('Action')
        # allow_action为空，表示没有设置允许的操作，则放行所有Action
        if allow_action is None or action in allow_action:
            t = request.get_data(as_text=True).split('=')
            t_dict = {}
            if len(t) > 1:
                for ix in range(0, len(t), 2):  # start=0, stop=len(t), step=2
                    t_dict[str(t[ix])] = t[ix + 1]
            args = dict(request.args)
            args.update(t_dict)
            args.update(request.form)
            if len(args) == 0 and use_action:
                global example
                if example is None:
                    example = "http://127.0.0.1:5000/?Action=CreatePerson&GroupName=test&PersonName=yk"
                return jsonify({"Success": True,
                                "Info": "服务正常，请使用Action传入指定的操作！",
                                "Example": example})
            logger.debug("执行操作{}".format(action))
            logger.debug(args)
            try:
                # result = eval("{}(args)".format(action))
                result = deal(args) or {}  # 需要确保result不为空
            except:
                traceback.print_exc()
                logger.debug(f"远程调用了{action}方法，但该方法在服务端没有定义！")
                return jsonify({"Success": False, "Info": "未知命令：{}".format(action)})
            result = name_transfer(result, 'CamelCase')  # 将字典中key的命名方式修改为驼峰命名方式
            result1 = jsonify(result)
            return result1

        else:
            logger.debug("操作{}不允许".format(action))
            return jsonify({"Success": False, "Info": "操作不允许，操作类型：{}".format(action)})

    @app.route('/download/<file_name>', methods=['GET'])
    @cross_origin(origins="*", methods=['GET'])
    def download_file(file_name):
        """
        下载文件功能

        前段通过 ip/download/<下载参数>进行下载，这里的下载参数一般是文件名。
        @download
        def deal(args_str):
            return os.path.join("C:/users", args_str)

        :param file_name: 前段传入的下载参数，一般为文件名
        :return:
        """
        directory = os.path.join(os.getcwd(), 'static')
        try:
            if download_func is not None:
                file_abs = download_func(file_name)
                if file_abs:
                    directory = os.path.dirname(file_abs)
                    file_name = os.path.basename(file_abs)
            response = make_response(send_from_directory(directory, file_name, as_attachment=True))
            return response
        except Exception as e:
            return jsonify({"Info": "文件下载失败！"})

    logger.info(f"服务已启动，通过http://localhost:{port}访问")
    from waitress import serve
    serve(app, host=host, port=port)

    return app


def download(func):
    """
    download装饰器，用于指定用户下载的文件
    用法示例：
    @download
    def transfer_file(args): # 这里args是前段通过 http://ip/download/<args>传递的下载参数
        return os.path.join("C:/static", args)  # 则前段会下载到服务器端 C:/static/args路径对应的函数

    :param func:
    :return:
    """
    global download_func
    download_func = func

    def wrap(*args, **kwargs):
        return func(*args, **kwargs)

    return wrap



import dll_file
import sys

dbp_api = dll_file.DllMode(ip="@ip@", user="@user@", passwd_str="@passwd@", port="@port@")

if len(sys.argv) == 1:
    print("缺少命令行参数，need cmd parameters")

json_str = sys.argv[1]

tags_values = eval(json_str)
tags = list(tags_values.keys())
values = list(tags_values.values())
dbp_api.write_snapshot_double(tags=tags, values=values)
print("写入SIS成功")



import copy
import traceback

import numpy as np

from yangke.sis.dll_file import init_dbp_api
from yangke.sis.sis_io import get_tag_value
from yangke.web.flaskserver import start_server_app, logger
from yangke.performance.tools.natural_gas import NaturalGas


def deal(args):
    action = args.get("Action").lower()
    result = eval("{}(args)".format(action))
    return result


def naturalgas(args):
    try:
        # 获取前端传入的数据
        compositions = NaturalGas.get_components_name()
        comp = {name: float(args.get(name) or 0) for name in compositions}
        ng = NaturalGas(comp)
        p = float(args.get("P") or 0.101325)
        t = float(args.get("T") or 20)
        t0 = float(args.get("T0") or 20)
        res = {
            "M": ng.mole_weight,
            "Hmg": ng.get_gcv_mass(t0) / 1000,
            "Hvg": ng.get_gcv_voln(20, 20) / 1000,
            "Hcg": ng.get_gcv_mole() / 1000,
            "Hmn": ng.get_ncv_mass(t0) / 1000,
            "Hvn": ng.get_ncv_voln(20, 20) / 1000,
            "Hcn": ng.get_ncv_mole() / 1000,
            "Z": ng.get_z_gas(p, t),
            "D": ng.density
        }
        return res
    except:
        traceback.print_exc()
        logger.debug("error")
        return {"Success": True,
                "Info": "服务正常，请使用Action传入指定的操作！",
                "Example": example}


class PolyCurve:
    def __init__(self, x_list, y_list, deg=1):
        """
        多项式拟合曲线，根据x_list和y_list拟合多项式曲线

        :param x_list: 已知的x点列表
        :param y_list: 一致的y点列表
        :param deg: 多项式阶次
        """
        self.x = np.array(x_list)
        self.y = np.array(y_list)
        self.z = np.polyfit(self.x, self.y, deg)
        self.func = np.poly1d(self.z)

    def get_value(self, x):
        """
        获取曲线上x点上对应的y值

        :param x:
        :return:
        """
        return self.func(x)

    def plot(self, xlim=None, ylim=None):
        import matplotlib.pyplot as plt
        xp = np.linspace(min(self.x), max(self.x))
        _ = plt.plot(self.x, self.y, '.', xp, self.func(xp), '-')
        if xlim is not None:
            plt.xlim()
        if ylim is not None:
            plt.ylim()
        plt.show()


def fit_curve(type="", power=None, flow_gas=None, a=-0.0693, b=1.3388, c=0.4832, d=3.7897):
    if type == "燃气流量vs纯凝功率":
        f_gas = [9.31, 7.5, 5.92, 4.54]
        power = [68.443, 53.504, 39.202, 27.098]
        _ = PolyCurve(x_list=power, y_list=f_gas, deg=3)
        _.plot()
        return _.func
    elif type == "纯凝功率vs燃气流量":
        f_gas = [9.31, 7.5, 5.92, 4.54]
        power = [68.443, 53.504, 39.202, 27.098]
        _ = PolyCurve(x_list=f_gas, y_list=power, deg=3)
        _.plot()


def get_history():
    from datetime import datetime, timedelta
    tags1 = {  # 可读参数，部分也可以写入，但不建议从该程序中写入
        "N1DCS.TCS110RCAOG_B120_01": "调压站天然气流量1",
        "N1DCS.TCS110RCAOG_B120_02": "调压站天然气流量2",
        "N1DCS.TCS110RCAOG_B116_01": "#1电功率",
        "N1DCS.TCS110RCAOG_B116_02": "#2电功率",
        "N1DCS.TCS110RCAOG_B116_03": "#3电功率",

    }
    now = datetime.now()
    p_list1 = []
    p_list2 = []
    p_list3 = []
    p_list4 = []
    api = init_dbp_api()
    his_value = api.get_his_value(tags=list(tags1.keys()), tags_description=list(tags1.values()),
                                  start_time=now - timedelta(days=0, hours=1),
                                  end_time=now - timedelta(days=0, hours=0), time_interval=60)
    his_value = his_value[his_value["#1电功率"] < 1]  # 剔除1号机运行时的数据
    his_value = his_value[his_value["#2电功率"] < 1]  # 剔除2号机运行时的数据
    his_value = his_value[his_value["#3电功率"] > 20]  # 剔除3号机停机时的数据
    his_value = his_value.set_index("DateTime").mean()

    his_value["天然气流量"] = (his_value["调压站天然气流量1"] + his_value["调压站天然气流量2"]) / 2

    data_file2 = []
    for i in range(12):
        api = init_dbp_api()
        his_value = api.get_his_value(tags=list(tags1.keys()), tags_description=list(tags1.values()),
                                      start_time=now - timedelta(days=30 * (i + 1), hours=0),
                                      end_time=now - timedelta(days=30 * i, hours=0), time_interval=60)
        his_value = his_value.set_index("DateTime").rolling(window=10).mean()
        his_value.dropna(how="any", inplace=True)
        # his_value_bak = copy.deepcopy(his_value)
        his_value.drop(his_value.tail(10).index, inplace=True)  # 删掉最后十行数据
        his_value = drop_revulsion(his_value, "#1电功率", 10, 10)  # 删除10分钟功率变化超过10MW的数据行
        his_value["大气压力"] = his_value["大气压力"] / 10  # 单位转换
        his_value["环境湿度"] = his_value["环境湿度"] / 100  # 单位转换

        his_value.to_csv(f"origin_{i}.csv")  # his_value为原始数据，不删除任何时间的数据


def start_server():
    # run()  # 启动定时执行
    heat_supply({})
    start_server_app(deal=deal, port=5000,
                     example_url=example)


if __name__ == "__main__":
    example = "http://127.0.0.1:5000/?Action=NaturalGas&CH4=0.98&N2=0.02&T=15&P=3.8&T0=20"
    start_server()



# -*- coding: utf-8 -*-
import datetime
import re
import sys

from selenium.common.exceptions import WebDriverException
from yangke.common.config import logger

import scrapy
from scrapy.linkextractors import LinkExtractor
from scrapy.selector import SelectorList
from scrapy.spiders import CrawlSpider, Rule
from pathlib2 import Path

from yangke.common.mysql import read_dataframe_from_mysql, connect_mysql, has_table
from selenium import webdriver

from stock10jqka.items import *

# noinspection NonAsciiCharacters
from stock10jqka.settings import mysql_config

print(f"__file__:{__file__}")
print(f"__package__:{__package__}")


class JqkaSpiderSpider(CrawlSpider):
    name = 'jqka_spider'
    allowed_domains = ['10jqka.com.cn']

    # 初始化mysql数据库
    config = mysql_config.copy()  # 浅拷贝
    try:
        # 建立mysql数据库连接
        connect_mysql(config=config, create_if_not_exist=True, use_pymysql=False, return_type="engine")
    except:
        raise Exception("初始化mysql时发生异常，检查mysql服务是否开启！")
    print("初始化mysql数据库成功！")

    # 如果不存在symbols表，需要初始化该表
    if not has_table("symbols"):
        import yangke.stock.main as get_symbols
        get_symbols.runCMD(f"python {get_symbols.__file__} --command getSymbols")

    codes_list = read_dataframe_from_mysql("symbols")['symbol']
    codes_list = list(codes_list)
    start_urls = []
    for code in codes_list:
        start_urls.append('http://stockpage.10jqka.com.cn/{}/'.format(code))
    # start_urls = ["http://stockpage.10jqka.com.cn/000001/"]
    del codes_list, code

    rules = (
        # 正则\d表示数字，$表示匹配结尾，即必须以\d结尾
        # 匹配 http://stockpage.10jqka.com.cn/600666/ 且不能匹配 http://stockpage.10jqka.com.cn/600666/event/
        # Rule(LinkExtractor(allow=r'http:\/\/stockpage\.10jqka.+\/\d+\/$',
        #                    restrict_xpaths="//ul[@class='news_list stat']"),  # 限制在公司新闻栏进行跟进
        #      # callback='parse_item',
        #      follow=True),
        Rule(LinkExtractor(allow=r'http:\/\/stockpage\.10jqka.+\/\d+\/$'),  # 在当前页面进行解析
             callback='parse_item',
             follow=False),
        # 匹配 http://stock.10jqka.com.cn/20200315/c618455360.shtml
        # Rule(LinkExtractor(allow=r'http:\/\/stock\.10jqka.+\/\d+\/.+\.shtml'), callback='parse_news',
        #        follow=False),  # 在跟进页面进行解析，这里cb_kwargs会传给callback指定的函数，Rule最终也要调用Request类
    )

    def __init__(self, *args, **kwargs):
        # 初始化爬虫使用的browser，必须安装selenium的edge/firefox/chrome中的任意一个
        try:
            self.browser = webdriver.Edge()
            logger.info("使用Edge浏览器进行数据爬取")
        except (FileNotFoundError, WebDriverException):
            try:
                self.browser = webdriver.Firefox()
                logger.info("使用Firefox浏览器进行数据爬取")
            except (FileNotFoundError, WebDriverException):
                try:
                    self.browser = webdriver.Chrome()
                    logger.info("使用Chrome浏览器进行数据爬取")
                except (FileNotFoundError, WebDriverException):
                    logger.error("selenium webdriver初始化失败！")
                    sys.exit(1)
        super(JqkaSpiderSpider, self).__init__()

    def parse_item(self, response):
        # noinspection NonAsciiCharacters
        公司简介 = response.xpath("//dl[@class='company_details']")  # 公司简介栏
        # noinspection NonAsciiCharacters
        牛叉诊股 = response.xpath("//div[@class='sub_cont_4 m_s_l']")  # 牛叉诊股栏
        # noinspection NonAsciiCharacters
        机构评级 = response.xpath("//div[@class='sub_cont_5']")  # 机构评级栏
        # print(len(公司简介), len(牛叉诊股), len(机构评级))
        # noinspection NonAsciiCharacters
        评分, 击败百分数, short_trend, medium_trend, long_trend, tech_trend, fund_trend, msg_trend, industry_trend, basic_trend = self.parse_牛叉诊股(
            牛叉诊股)
        # noinspection NonAsciiCharacters
        评级 = self.parse_机构评级(机构评级)  # 0,1,2,3,4,5对应【没有数据, 卖出, 减持, 中性, 增持, 买入】
        date = datetime.datetime.today()
        # noinspection NonAsciiCharacters
        所属地域, 设计概念, 主营业务, 上市日期, 每股净资产, 每股收益, 净利润, 净利润增长率, 营业收入, 每股现金流, 每股公积金, 每股未分配利润, 总股本, 流通股 = self.parse_公司简介(公司简介)

        temp = response.xpath("//h1[@class='m_logo fl']")[0]
        temp = temp.xpath(".//strong/text()").getall()
        name = temp[0]
        code = temp[1]

        item1 = StockInfoItem(code=code, name=name, location=所属地域, concept=设计概念, business=主营业务,
                              date_listing=上市日期)
        item2 = Stock10JqkaItem(code=code, date=date.date(), 诊股_评分=评分, 诊股_击败=击败百分数, 诊股_短期趋势=short_trend,
                                诊股_中期趋势=medium_trend, 诊股_长期趋势=long_trend, 诊股_技术面=tech_trend,
                                诊股_资金面=fund_trend, 诊股_消息面=msg_trend, 诊股_行业面=industry_trend,
                                诊股_基本面=basic_trend, 评级=评级, 设计概念=设计概念, 主营业务=主营业务, 地域=所属地域,
                                每股净资产=每股净资产, 每股收益=每股收益, 净利润=净利润, 净利润增长率=净利润增长率,
                                营业收入=营业收入, 每股现金流=每股现金流, 每股公积金=每股公积金,
                                每股未分配利润=每股未分配利润, 总股本=总股本, 流通股=流通股)
        yield item1
        yield item2
        # noinspection NonAsciiCharacters
        公司新闻 = response.xpath("//ul[@stat='f10_spqk_gsxw']//@href").getall()
        pattern_news = re.compile(r'http:\/\/stock\.10jqka.+\/\d+\/.+\.shtml')
        urls = [url for url in 公司新闻 if pattern_news.match(url)]

        for href in urls:
            print("GET {}".format(href))
            yield scrapy.Request(url=href, callback=self.parse_news, cb_kwargs={"code": code})

    # noinspection All
    def parse_公司简介(self, selectors: SelectorList):
        dds = selectors[0].xpath(".//dd")
        所属地域 = dds[0].xpath(".//text()").get()
        设计概念 = dds[1].xpath(".//@title").get()
        主营业务 = dds[3].xpath(".//@title").get()
        # 去除括号内的备注内容，删除最后的句号
        主营业务 = 主营业务.strip()
        上市日期 = dds[4].xpath(".//text()").get()
        # 以下为与日期有关的
        # 数字的正则表达式：(-?\d+\.\d+)
        # python中正则表达式中的小括号有特殊的作用，小括号括起来的部分为re.findall()返回的结果，而整个正则表达式决定是不是匹配
        # 如 re.findall('(\d+)%?', '43%')[0] -> 43
        # 而 re.findall('(\d+%?)', '43%')[0] -> 43%
        # 如果没有小括号，则相当于整个正则表达式都在小括号内部
        # 如果有多个小括号，都会返回成tuple，每个小括号内的匹配结果都会成为tuple项
        # 如 re.findall('(\d+)\.(\d+)', '43.23%') -> [('43', '23')]
        float_re = r'(-?\d+\.\d+)'
        character_re = '[\u4e00-\u9fa5]+'  # 汉字的正则表达式，用于校验单位是否不是正常单位

        section = dds[5].xpath(".//text()").get()
        每股净资产 = re.findall(float_re, section)[0]  # 元
        unit = re.findall(character_re, section)[0]  # 元
        assert unit == "元", "网页上爬取的【每股净资产】单位与软件数据库中单位不一致，请检查..."

        section = dds[6].xpath(".//text()").get()
        每股收益 = re.findall(float_re, section)[0]  # 元
        unit = re.findall(character_re, section)[0]  # 元
        assert unit == "元", "网页上爬取的【每股收益】单位与软件数据库中单位不一致，请检查..."

        section = dds[7].xpath(".//text()").get()
        净利润 = re.findall(float_re, section)[0]  # 亿元
        unit = re.findall(character_re, section)[0]  # 亿元
        assert unit == "亿元", "网页上爬取的【净利润】单位与软件数据库中单位不一致，请检查..."

        section = dds[8].xpath(".//text()").get()
        净利润增长率 = re.findall(float_re, section)[0]  #

        section = dds[9].xpath(".//text()").get()
        营业收入 = re.findall(float_re, section)[0]  # 亿元
        unit = re.findall(character_re, section)[0]  # 亿元
        assert unit == "亿元", "网页上爬取的【营业收入】单位与软件数据库中单位不一致，请检查..."

        section = dds[10].xpath(".//text()").get()
        每股现金流 = re.findall(float_re, section)[0]  # 元
        unit = re.findall(character_re, section)[0]  # 元
        assert unit == "元", "网页上爬取的【每股现金流】单位与软件数据库中单位不一致，请检查..."

        section = dds[11].xpath(".//text()").get()
        每股公积金 = re.findall(float_re, section)[0]  # 元
        unit = re.findall(character_re, section)[0]  # 元
        assert unit == "元", "网页上爬取的【每股公积金】单位与软件数据库中单位不一致，请检查..."

        section = dds[12].xpath(".//text()").get()
        每股未分配利润 = re.findall(float_re, section)[0]  # 元
        unit = re.findall(character_re, section)[0]  # 元
        assert unit == "元", "网页上爬取的【每股未分配利润】单位与软件数据库中单位不一致，请检查..."

        section = dds[13].xpath(".//text()").get()
        总股本 = re.findall(float_re, section)[0]  # 亿
        unit = re.findall(character_re, section)[0]  # 亿
        assert unit == "亿", "网页上爬取的【总股本】单位与软件数据库中单位不一致，请检查..."

        section = dds[14].xpath(".//text()").get()
        流通股 = re.findall(float_re, section)[0]  # 亿
        unit = re.findall(character_re, section)[0]  # 亿
        assert unit == "亿", "网页上爬取的【流通股】单位与软件数据库中单位不一致，请检查..."
        return 所属地域, 设计概念, 主营业务, 上市日期, 每股净资产, 每股收益, 净利润, 净利润增长率, 营业收入, 每股现金流, 每股公积金, 每股未分配利润, 总股本, 流通股

    # noinspection All
    def parse_news(self, response, **cb_kwargs):
        code = cb_kwargs.get('code')
        content_div = response.xpath("//div[@class='main-text atc-content']")
        # content_list = content_div.xpath(".//p/text()").getall()  # 只能从<p>节点获得直接的内容
        content_list = content_div.xpath(".//p").xpath('string(.)').getall()  # 可以获得<p><strong>内容</><>
        content_list = [p.strip() for p in content_list if p.strip() != ""]

        # 去掉新闻后的网页宣传后缀
        # post = content_div.xpath(".//a[@target='_blank']").xpath('string(.)').getall()[-2].strip()
        # content_list = content_list[:-2]

        content = "\n".join(content_list)
        title = response.xpath("//h2[@class='main-title']/text()").get()
        temp_section = response.xpath("//div[@class='info-fl fl']")
        date = temp_section.xpath(".//span[@id='pubtime_baidu']/text()").get()
        # source=temp_section.xpath(".//span[@") # 新闻来源可能是个图片，暂时不爬取来源
        item = NewsItem(code=code, 标题=title, date=date, 内容=content, url=response.url)
        yield item

    # noinspection All
    def parse_牛叉诊股(self, sections: SelectorList):
        for selector in sections:
            评分 = selector.xpath(".//span[@class='analyze-num']/text()").get()
            击败百分数 = selector.xpath(".//span[@class='analyze-tips']/text()").get()
            击败百分数 = re.findall(r"(\d+)", 击败百分数)[0]
            trend = selector.xpath(".//div[@class='txt-main']/text()").getall()
            short_trend, medium_trend, long_trend = trend[0], trend[1], trend[2]
            诊股 = selector.xpath(".//div[@class='analyze-stars']/i/text()").getall()
            tech_trend, fund_trend, msg_trend, industry_trend, basic_trend \
                = 诊股[0], 诊股[1], 诊股[2], 诊股[3], 诊股[4]
            result = (评分, 击败百分数, short_trend, medium_trend, long_trend,
                      tech_trend, fund_trend, msg_trend, industry_trend, basic_trend)
        return result

    # noinspection All
    def parse_机构评级(self, sections: SelectorList):
        for selector in sections:
            评级 = selector.xpath(".//div[@class='jg-level']/span/@class").get()
            评级 = re.findall(r"\d+", 评级)[0]

        return 评级

    def close(self, reason):
        # 关闭selenium打开的浏览器
        self.browser.close()
        super(JqkaSpiderSpider, self).close(spider=self, reason="all done")
